from .cv import Term, TermSet


class SoftwareName(Term):
    pass


class Software(object):

    @classmethod
    def is_name(cls, name):
        return name in software_names_by_name

    def __init__(self, name=None, id=None, version=None, **options):
        if name is None:
            name, options = self._resolve_name_from_kwargs(options)
        if name is None and id is not None:
            name = id
        self.name = name
        self.id = id or name
        self.version = version
        self.options = options

    def _resolve_name_from_kwargs(self, options):
        names = dict()
        not_names = dict()
        for key, value in options.items():
            if self.is_name(key):
                names[key] = value
            else:
                not_names[key] = value
        options = not_names
        if len(names) == 1:
            name = list(names.keys())[0]
        elif len(names) == 0:
            name = None
        else:
            raise ValueError("Multiple possible names found")
        return name, options

    def __str__(self):
        return self.name

    def __repr__(self):
        template = '{self.__class__.__name__}({self.name}, {self.id}, {self.version})'
        return template.format(self=self)

    def __eq__(self, other):
        try:
            return self.name == other.name
        except AttributeError:
            return str(self) == str(other)


software_names = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('software', list_name='software_names', term_cls_name="SoftwareName", writer=cog.out)
# ]]]
software_names = TermSet([
    SoftwareName(u'SCiLS software', u'MS:1002383',
                 (u'SCiLS software for data acquisition and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'acquisition software', u'MS:1001455',
                 (u'Acquisition software.'),
                 'software',
                 [u'software']),
    SoftwareName(u'data processing software', u'MS:1001457',
                 (u'Data processing software.'),
                 'software',
                 [u'software']),
    SoftwareName(u'analysis software', u'MS:1001456',
                 (u'Analysis software.'),
                 'software',
                 [u'software']),
    SoftwareName(u'SCIEX software', u'MS:1000690',
                 (u'SCIEX or Applied Biosystems software for data acquisition'
                  u'and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'Applied Biosystems software', u'MS:1000691',
                 (u'Applied Biosystems|MDS SCIEX software for data acquisition'
                  u'and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'Bruker software', u'MS:1000692',
                 (u'Bruker software for data acquisition and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'Thermo Finnigan software', u'MS:1000693',
                 (u'Thermo Finnigan software for data acquisition and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'Waters software', u'MS:1000694',
                 (u'Waters software for data acquisition and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'SRM software', u'MS:1000871',
                 (u'Software used to predict, select, or optimize transitions or'
                  u'analyze the results of selected reaction monitoring runs.'),
                 'software',
                 [u'software']),
    SoftwareName(u'peptide attribute calculation software', u'MS:1000873',
                 (u'Software used to predict or calculate numerical attributes'
                  u'of peptides.'),
                 'software',
                 [u'software']),
    SoftwareName(u'Agilent software', u'MS:1000689',
                 (u'Agilent software for data acquisition and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'quantitation software name', u'MS:1001139',
                 (u'Quantitation software name.'),
                 'software',
                 [u'software', u'quantification information']),
    SoftwareName(u'LECO software', u'MS:1001798',
                 (u'LECO software for data acquisition and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'custom unreleased software tool', u'MS:1000799',
                 (u'A software tool that has not yet been released. The value'
                  u'should describe the software. Please do not use this term'
                  u'for publicly available software - contact the PSI-MS working'
                  u'group in order to have another CV term added.'),
                 'software',
                 [u'software']),
    SoftwareName(u'BSI software', u'MS:1001949',
                 (u'Bioinformatics Solutions Inc. Software for data processing'
                  u'and analysis.'),
                 'software',
                 [u'software']),
    SoftwareName(u'Shimadzu Corporation software', u'MS:1001557',
                 (u'Shimadzu Corporation software.'),
                 'software',
                 [u'software']),
    SoftwareName(u'SCiLS Lab', u'MS:1002384',
                 (u'SCiLS Lab software.'),
                 'software',
                 [u'SCiLS software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Analyst', u'MS:1000551',
                 (u'SCIEX or Applied Biosystems|MDS SCIEX software for data'
                  u'acquisition.'),
                 'software',
                 [u'SCIEX software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'apexControl', u'MS:1000706',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'MALDI Solutions LC-MALDI', u'MS:1002381',
                 (u'Software for automated LC-MALDI analysis and reporting.'),
                 'software',
                 [u'acquisition software', u'analysis software', u'data processing software', u'Shimadzu Corporation software', u'software']),
    SoftwareName(u'dpControl', u'MS:1000720',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'esquireControl', u'MS:1000721',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'HCTcontrol', u'MS:1000725',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'micrOTOFcontrol', u'MS:1000726',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'spControl', u'MS:1000737',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'ChromaTOF HRT software', u'MS:1001877',
                 (u'Software for acquisition, processing and analysis of data'
                  u'for LECO instruments.'),
                 'software',
                 [u'acquisition software', u'analysis software', u'data processing software', u'LECO software', u'software']),
    SoftwareName(u'6300 Series Ion Trap Data Analysis Software', u'MS:1000688',
                 (u'Software for data analysis of 6300 series ion trap mass'
                  u'spectrometers.'),
                 'software',
                 [u'Agilent software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ChromaTOF software', u'MS:1001799',
                 (u'Software for acquisition, processing and analysis of data'
                  u'for LECO instruments.'),
                 'software',
                 [u'acquisition software', u'analysis software', u'data processing software', u'LECO software', u'software']),
    SoftwareName(u'MassHunter Data Acquisition', u'MS:1000678',
                 (u'Software for data acquisition of 6000 series instruments.'),
                 'software',
                 [u'Agilent software', u'acquisition software', u'software']),
    SoftwareName(u'MassHunter Easy Access', u'MS:1000679',
                 (u'Software for open access data acquisition.'),
                 'software',
                 [u'Agilent software', u'acquisition software', u'software']),
    SoftwareName(u'GPS Explorer', u'MS:1000661',
                 (u'SCIEX or Applied Biosystems software for data acquisition'
                  u'and analysis.'),
                 'software',
                 [u'SCIEX software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Voyager Biospectrometry Workstation System', u'MS:1000539',
                 (u'Applied Biosystems MALDI-TOF data acquisition and analysis'
                  u'system.'),
                 'software',
                 [u'Applied Biosystems software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Xcalibur', u'MS:1000532',
                 (u'Thermo Finnigan software for data acquisition and analysis.'),
                 'software',
                 [u'Thermo Finnigan software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MassLynx', u'MS:1000534',
                 (u'Micromass software for data acquisition and analysis.'),
                 'software',
                 [u'Waters software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'4700 Explorer', u'MS:1000537',
                 (u'Applied Biosystems software for data acquisition and'
                  u'analysis.'),
                 'software',
                 [u'Applied Biosystems software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Data Explorer', u'MS:1000536',
                 (u'Applied Biosystems software for data acquisition and'
                  u'analysis.'),
                 'software',
                 [u'Applied Biosystems software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'4000 Series Explorer Software', u'MS:1000659',
                 (u'SCIEX or Applied Biosystems software for data acquisition'
                  u'and analysis.'),
                 'software',
                 [u'SCIEX software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'FlexControl', u'MS:1000540',
                 (u'Bruker software for data acquisition.'),
                 'software',
                 [u'Bruker software', u'acquisition software', u'software']),
    SoftwareName(u'SCIEX TOF/TOF Series Explorer Software', u'MS:1001483',
                 (u'SCIEX or Applied Biosystems software for TOF/TOF data'
                  u'acquisition and analysis.'),
                 'software',
                 [u'SCIEX software', u'acquisition software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MALDI Solutions', u'MS:1001558',
                 (u'Shimadzu Biotech software for data acquisition, processing,'
                  u'and analysis.'),
                 'software',
                 [u'acquisition software', u'analysis software', u'data processing software', u'Shimadzu Corporation software', u'software']),
    SoftwareName(u'Lipid-Pro', u'MS:1002972',
                 (u'A computational lipid identification solution for untargeted'
                  u'lipidomics on data-independent acquisition tandem mass'
                  u'spectrometry platforms.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'Trapper', u'MS:1000553',
                 (u'A software program for converting Agilent MassHunter format'
                  u'to mzXML or mzML. Trapper was originally developed at the'
                  u'Institute for Systems Biology.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'CLINPROT', u'MS:1000708',
                 (u'Bruker CLINPROT software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'CLINPROT micro', u'MS:1000709',
                 (u'Bruker CLINPROT micro software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'BioTools', u'MS:1000707',
                 (u'Bruker software for data analysis.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'LipidMatch', u'MS:1002969',
                 (u'An automated workflow for rule-based lipid identification'
                  u'using untargeted high-resolution tandem mass spectrometry'
                  u'data.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'preprocessing software', u'MS:1002386',
                 (u'Preprocessing software.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'PEAKS Online', u'MS:1001947',
                 (u'PEAKS Online software for high throughput data analysis.'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'data processing software', u'software', u'quantification information']),
    SoftwareName(u'PEAKS Studio', u'MS:1001946',
                 (u'PEAKS Studio software for data analysis.'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'data processing software', u'software', u'quantification information']),
    SoftwareName(u'DataAnalysis', u'MS:1000719',
                 (u'Bruker software for data analysis.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'CompassXtract', u'MS:1000718',
                 (u'Bruker software library for data access.'),
                 'software',
                 [u'Bruker software', u'data processing software', u'software']),
    SoftwareName(u'Compass OpenAccess', u'MS:1000715',
                 (u'Bruker compass OpenAccess software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Compass for micrOTOF', u'MS:1000714',
                 (u'Bruker Compass for micrOTOF software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'CompassXport', u'MS:1000717',
                 (u'Bruker stand-alone software for data conversion.'),
                 'software',
                 [u'Bruker software', u'data processing software', u'software']),
    SoftwareName(u'Compass for HCT/esquire', u'MS:1000713',
                 (u'Bruker Compass for HCT/esquire software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Compass', u'MS:1000712',
                 (u'Bruker Compass software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ALEX123', u'MS:1002977',
                 (u'Analysis of lipid experiments 123, a calculator with m/z'
                  u'values of intact lipid molecules (MS1) and their fragment'
                  u'ions at the MS2 and MS3 level.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'ALEX', u'MS:1002976',
                 (u'Analysis of lipid experiments, a calculator for m/z values'
                  u'of intact lipid molecules (MS1).'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'LipidFinder', u'MS:1002973',
                 (u'A computational workflow for the discovery of lipids for the'
                  u'identification of eicosanoid-phosphoinositides in platelets.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'LipidBlast', u'MS:1002971',
                 (u'LC-MS-based lipidomics and automated identification of'
                  u'lipids using the LipidBlast in-silico MS/MS library.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'Greazy', u'MS:1002970',
                 (u'Open-source software for automated phospholipid tandem mass'
                  u'spectrometry identification.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'LipidQA', u'MS:1002980',
                 (u'Lipid qualitative/quantitative analysis software for'
                  u'identification and quantitation of complex lipid molecular'
                  u'species.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'flexImaging', u'MS:1000722',
                 (u'Bruker software for data analysis.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProfileAnalysis', u'MS:1000728',
                 (u'Bruker software for data analysis.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MSDK', u'MS:1002645',
                 (u'Mass Spectrometry Development Kit (MSDK) is a Java library'
                  u'of algorithms for processing of mass spectrometry data."'
                  u'[PSI:PI'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'QuantAnalysis', u'MS:1000736',
                 (u'Bruker software for data analysis.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'LipidXplorer', u'MS:1002968',
                 (u'Software for consensual cross-platform lipidomics.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'MzWiff', u'MS:1000591',
                 (u'A software program for converting Applied Biosystems wiff'
                  u'file format to the mzXML or mzML format. MzWiff is currently'
                  u'maintained at the Institute for Systems Biology. It replaces'
                  u'the slower mzStar program.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'LipidHunter', u'MS:1002967',
                 (u'Software for identification of phospholipids by high-'
                  u'throughput processing of LC-MS and shotgun lipidomics'
                  u'datasets.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'Lipid Data Analyzer', u'MS:1002965',
                 (u'Lipid Data Analyzer software for lipid quantification.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'SQID', u'MS:1001886',
                 (u'Software for data analysis of peptides and proteins.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Maltcms', u'MS:1002344',
                 (u'Modular Application Toolkit for Chromatography Mass-'
                  u'Spectrometry is an application framework mainly for'
                  u'developers." [PSI:PI'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MZmine', u'MS:1002342',
                 (u'A framework for differential analysis of mass spectrometry'
                  u'data." [PMID:16403790'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'PepFinder', u'MS:1002524',
                 (u'Thermo Scientific PepFinder BioPharma analysis software.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'Spectrum Mill for MassHunter Workstation', u'MS:1000687',
                 (u'Software for protein identification and characterization of'
                  u'complex protein digest mixtures.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'METLIN', u'MS:1000686',
                 (u'Personal Metabolite Database for MassHunter Workstation.'
                  u'Software for identification of human metabolites.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MassHunter Mass Profiler', u'MS:1000685',
                 (u'Software for quantitation and statistical analysis of TOF'
                  u'and Q-TOF LC/MS data.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Genespring MS', u'MS:1000684',
                 (u'Software for quantitation and statistical analysis of TOF'
                  u'and Q-TOF LC/MS data.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MassHunter BioConfirm', u'MS:1000683',
                 (u'Software for protein characterization.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MassHunter Metabolite ID', u'MS:1000682',
                 (u'Software for identification of metabolites.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MassHunter Quantitative Analysis', u'MS:1000681',
                 (u'Software for quantitation of Triple Quadrupole and'
                  u'Quadrupole Time-of-Flight data.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MassHunter Qualitative Analysis', u'MS:1000680',
                 (u'Software for data analysis of data from 6000 series'
                  u'instruments.'),
                 'software',
                 [u'Agilent software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'postprocessing software', u'MS:1002414',
                 (u'Postprocessing software.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'XCMS', u'MS:1001582',
                 (u'Bioconductor package XCMS for preprocessing high-throughput,'
                  u'untargeted analyte profiling data.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'UNIFY', u'MS:1001796',
                 (u'Waters UNIFY software for liquid chromatography and mass'
                  u'spectrometry acquisition.'),
                 'software',
                 [u'Waters software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Empower', u'MS:1001795',
                 (u'Waters Empower software for liquid chromatography and mass'
                  u'spectrometry acquisition.'),
                 'software',
                 [u'Waters software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Pro Quant', u'MS:1000670',
                 (u'Applied Biosystems|MDS SCIEX software for protein ID and'
                  u'quant by iTRAQ.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Pro BLAST', u'MS:1000671',
                 (u'Applied Biosystems|MDS SCIEX software for MS-BLAST'
                  u'identification.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MultiQuant', u'MS:1000674',
                 (u'Applied Biosystems|MDS SCIEX software for MRM-based'
                  u'quantitation.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteinPilot Software', u'MS:1000663',
                 (u'SCIEX or Applied Biosystems|MDS SCIEX software for protein'
                  u'ID and quant.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'LightSight Software', u'MS:1000662',
                 (u'SCIEX or Applied Biosystems|MDS SCIEX software metabolite'
                  u'identification.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MarkerView Software', u'MS:1000665',
                 (u'Applied Biosystems|MDS SCIEX software for metabolomics and'
                  u'biomarker profiling.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TissueView Software', u'MS:1000664',
                 (u'Applied Biosystems|MDS SCIEX software for tissue imaging.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'BioAnalyst', u'MS:1000667',
                 (u'Applied Biosystems|MDS SCIEX software for bio-related data'
                  u'exploration.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MRMPilot Software', u'MS:1000666',
                 (u'Applied Biosystems|MDS SCIEX software for MRM assay'
                  u'development.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Pro ICAT', u'MS:1000669',
                 (u'Applied Biosystems|MDS SCIEX software for protein ID and'
                  u'quant by ICAT.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Pro ID', u'MS:1000668',
                 (u'Applied Biosystems|MDS SCIEX software for protein'
                  u'identification.'),
                 'software',
                 [u'SCIEX software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'conversion software', u'MS:1002333',
                 (u'Computer software primarily designed to convert data'
                  u'represented in one format to another format, sometimes with'
                  u'minor data alterations in the process.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'ms_deisotope', u'MS:1002990',
                 (u'ms_deisotope, a library for deisotoping and charge state'
                  u'deconvolution of mass spectra.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'PEAKS Node', u'MS:1001948',
                 (u'PEAKS Node software for high throughput data analysis.'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'data processing software', u'software', u'quantification information']),
    SoftwareName(u'massWolf', u'MS:1000538',
                 (u'A software for converting Waters raw directory format to'
                  u'mzXML or mzML. MassWolf was originally developed at the'
                  u'Institute for Systems Biology.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'Bioworks', u'MS:1000533',
                 (u'Thermo Finnigan software for data analysis of peptides and'
                  u'proteins.'),
                 'software',
                 [u'Thermo Finnigan software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'FlexAnalysis', u'MS:1000535',
                 (u'Bruker software for data analysis.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Proteome Discoverer', u'MS:1000650',
                 (u'Thermo Scientific software for data analysis of peptides and'
                  u'proteins.'),
                 'software',
                 [u'Thermo Finnigan software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MSnbase', u'MS:1002870',
                 (u'Bioconductor package MSnbase provides infrastructure for'
                  u'manipulation, processing and visualization of mass'
                  u'spectrometry and proteomics data, ranging from raw to'
                  u'quantitative and annotated data.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'CAMERA', u'MS:1002871',
                 (u'Bioconductor package CAMERA for annotation of peak lists'
                  u'generated by xcms, rule based annotation of isotopes and'
                  u'adducts, isotope validation, EIC correlation based tagging'
                  u'of unknown adducts and fragments.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'mzR', u'MS:1002869',
                 (u'Bioconductor package mzR for reading and writing mass'
                  u'spectrometry data files.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'LOBSTAHS', u'MS:1002979',
                 (u'Adduct-Based lipidomics software for the discovery and'
                  u'identification of oxidative stress biomarkers.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'LIMSA', u'MS:1002978',
                 (u'Software tool for the quantitative analysis of mass'
                  u'spectrometric lipidome data.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'LipiDex', u'MS:1002974',
                 (u'An integrated software package for high-confidence lipid'
                  u'identification.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'pymzML', u'MS:1001914',
                 (u'Python module to interface mzML Data.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'ReAdW', u'MS:1000541',
                 (u'A software program for converting Thermo Finnigan RAW file'
                  u'format to mzXML or mzML. ReAdW was originally developed at'
                  u'the Institute for Systems Biology. Its whimsical interleaved'
                  u'spelling and capitalization is pronounced \\"readraw\\".'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'MzStar', u'MS:1000542',
                 (u'A software program for converting Applied Biosystems wiff'
                  u'file format to mzXML format. MzStar was originally developed'
                  u'at the Institute for Systems Biology. It is now obsoleted by'
                  u'the MzWiff program.'),
                 'software',
                 [u'data processing software', u'software']),
    SoftwareName(u'Maui', u'MS:1002452',
                 (u'The Maltcms Graphical User Interface." [PSI:PI'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP software', u'MS:1000752',
                 (u'TOPP (The OpenMS proteomics pipeline) software.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteoWizard software', u'MS:1000615',
                 (u'ProteoWizard software for data processing and analysis.'
                  u'Primarily developed by the labs of P. Malick and D. Tabb.'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'PinPoint', u'MS:1001912',
                 (u'Thermo Scientific PinPoint SRM analysis software.'),
                 'software',
                 [u'Thermo Finnigan software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteinLynx Global Server', u'MS:1000601',
                 (u'Waters software for data analysis.'),
                 'software',
                 [u'Waters software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Proteios', u'MS:1000600',
                 (u'Database application and analysis platform for proteomics."'
                  u'[PSI:MS'),
                 'software',
                 [u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'LIQUID', u'MS:1002975',
                 (u'An-open source software for identifying lipids in LC-MS/MS-'
                  u'based lipidomics data.'),
                 'software',
                 [u'lipidomics analysis software', u'data processing software', u'small molecule analysis software', u'software', u'analysis software']),
    SoftwareName(u'PIA', u'MS:1002387',
                 (u'PIA - Protein Inference Algorithms, a toolbox for protein'
                  u'inference and identification analysis." [PSI:PI'),
                 'software',
                 [u'postprocessing software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'PepNovo', u'MS:1002982',
                 (u'PepNovo tool for de novo peptide sequencing.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'pNovo', u'MS:1002983',
                 (u'pNovo tool for de novo peptide sequencing and identification'
                  u'using HCD spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Tide', u'MS:1002575',
                 (u'Tide open-source sequence search program developed at the'
                  u'University of Washington.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Morpheus', u'MS:1002661',
                 (u'Morpheus search engine.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'IdentiPy', u'MS:1002987',
                 (u'IdentiPy." [PMID:29682971'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Mascot Distiller', u'MS:1001488',
                 (u'Mascot Distiller.'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'software', u'quantification information']),
    SoftwareName(u'IsobariQ', u'MS:1002210',
                 (u'A quantitative software package designed for analysis of'
                  u'IPTL, TMT and iTRAQ data." [PMID:21067241,'
                  u'DOI:10.1021/pr1009977'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'software', u'quantification information']),
    SoftwareName(u'Ascore software', u'MS:1001984',
                 (u'Ascore software.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'ProteinScape', u'MS:1000734',
                 (u'Bruker ProteinScape software.'),
                 'software',
                 [u'Bruker software', u'analysis software', u'software']),
    SoftwareName(u'greylag', u'MS:1001461',
                 (u'Greylag identification software.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Mascot Parser', u'MS:1001478',
                 (u'Mascot Parser was used to analyze the spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'SpectraST', u'MS:1001477',
                 (u'SpectraST was used to analyze the spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'X\\!Tandem', u'MS:1001476',
                 (u'X!Tandem was used to analyze the spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'OMSSA', u'MS:1001475',
                 (u'Open Mass Spectrometry Search Algorithm was used to analyze'
                  u'the spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'mzidLib', u'MS:1002237',
                 (u'A library of Java routines for manipulating mzIdentML files.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Pepitome', u'MS:1001588',
                 (u'Tabb Lab software for spectral library searches on tandem'
                  u'mass spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MaxQuant', u'MS:1001583',
                 (u'MaxQuant is a quantitative proteomics software package'
                  u'designed for analyzing large mass spectrometric data sets.'
                  u'It is specifically aimed at high resolution MS data.'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'software', u'quantification information']),
    SoftwareName(u'Comet', u'MS:1002251',
                 (u'Comet open-source sequence search engine developed at the'
                  u'University of Washington.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Andromeda', u'MS:1002337',
                 (u'Andromeda is a peptide search engine.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Amanda', u'MS:1002336',
                 (u'Amanda scoring system for PSM identification.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'TopFD', u'MS:1002902',
                 (u'Top-down mass spectral feature detection.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'SEQUEST', u'MS:1001208',
                 (u'The name of the SEQUEST search engine.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Phenyx', u'MS:1001209',
                 (u'The name of the Phenyx search engine.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Mascot', u'MS:1001207',
                 (u'The name of the Mascot search engine.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Percolator', u'MS:1001490',
                 (u'Percolator.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'small molecule analysis software', u'MS:1002878',
                 (u'Software for the analysis of small molecules.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'TopMG', u'MS:1002903',
                 (u'A mass graph-based approach for the identification of'
                  u'modified proteoforms using top-down tandem mass spectra."'
                  u'[PMID:28453668'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'TopPIC', u'MS:1002901',
                 (u'TopPIC: a software tool for top-down mass spectrometry-based'
                  u'proteoform identification and characterization."'
                  u'[PMID:27423895'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'ProteinExtractor', u'MS:1001487',
                 (u'An algorithm for protein determination/assembly integrated'
                  u"into Bruker's ProteinScape."),
                 'software',
                 [u'Bruker software', u'analysis software', u'software']),
    SoftwareName(u'Mascot Integra', u'MS:1001489',
                 (u'Mascot Integra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Byonic', u'MS:1002261',
                 (u'Byonic search engine from Protein Metrics.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'PeptideShaker', u'MS:1002458',
                 (u'PeptideShaker is a software for the interpretation of'
                  u'proteomics identification results." [PSI:PI'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'TagRecon', u'MS:1001587',
                 (u'Tabb Lab software for reconciling sequence tags to a protein'
                  u'database.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'DirecTag', u'MS:1001586',
                 (u'Tabb Lab software for generating sequence tags from tandem'
                  u'mass spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MyriMatch', u'MS:1001585',
                 (u'Tabb Lab software for directly comparing peptides in a'
                  u'database to tandem mass spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'PAnalyzer', u'MS:1002076',
                 (u'PAnalyzer software for getting protein evidence categories.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'NIST MSPepSearch', u'MS:1002750',
                 (u'Search tool of the NIST (National Institute of Standards and'
                  u'Technology) for spectral library searches.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Trans-Proteomic Pipeline', u'MS:1002285',
                 (u'A suite of open source tools for the processing of MS2'
                  u'proteomics data developed by the Seattle Proteome Center at'
                  u'the Institute for Systems Biology.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Trans-Proteomic Pipeline software', u'MS:1002286',
                 (u'A software program that is a component of the Trans-'
                  u'Proteomic Pipeline.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'DTASelect', u'MS:1002598',
                 (u'Analysis software designed to reassemble the SEQUEST peptide'
                  u'identifications and to highlight the most significant'
                  u'matches." [PMID:12643522'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MSQuant', u'MS:1001977',
                 (u'MSQuant software.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'DeBunker', u'MS:1001973',
                 (u'DeBunker software.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Proline', u'MS:1002981',
                 (u'The Proline software suite for mass spectrometry based'
                  u'proteomics.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Novor', u'MS:1002984',
                 (u'Novor real-time peptide de novo sequencing software tool.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Scaffold', u'MS:1001561',
                 (u'Scaffold analysis software.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'ProLuCID', u'MS:1002596',
                 (u'The SEQUEST-like sequence search engine ProLuCID, developed'
                  u'in the Yates Lab at the Scripps Research Institute.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'xiFDR', u'MS:1002543',
                 (u'Target/Decoy based FDR estimation for cross-linking peptide-'
                  u'identifications.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Skyline mzQuantML converter', u'MS:1002546',
                 (u'A software package to convert Skyline report to mzQuantML."'
                  u'[PSI:PI'),
                 'software',
                 [u'quantitation software name', u'analysis software', u'software', u'quantification information']),
    SoftwareName(u'xi', u'MS:1002544',
                 (u'Search engine for cross-linked peptides.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MSPathFinder', u'MS:1002720',
                 (u'PNNL top-down/bottom-up analysis software for identifying'
                  u'peptides and proteoforms in fragmentation mass spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MetaMorpheus', u'MS:1002826',
                 (u'MetaMorpheus search engine.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MS-GF', u'MS:1002047',
                 (u'MS-GF software used to re-score the peptide-spectrum'
                  u'matches." [DOI:10.1074/mcp.M110.003731'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'ProteinProspector', u'MS:1002043',
                 (u'ProteinProspector software for data acquisition and'
                  u'analysis.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'MS-GF+', u'MS:1002048',
                 (u'MS-GF+ software used to analyze the spectra.'),
                 'software',
                 [u'analysis software', u'software']),
    SoftwareName(u'Cliquid', u'MS:1000672',
                 (u'SCIEX Cliquid software for data analysis and quantitation.'),
                 'software',
                 [u'SCIEX software', u'software']),
    SoftwareName(u'MIDAS Workflow Designer', u'MS:1000673',
                 (u'Applied Biosystems|MDS SCIEX software for MRM assay'
                  u'development.'),
                 'software',
                 [u'SCIEX software', u'software']),
    SoftwareName(u'Compass Security Pack', u'MS:1000716',
                 (u'Bruker compass Security Pack software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'ClinProTools', u'MS:1000711',
                 (u'Bruker ClinProTools software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'CLINPROT robot', u'MS:1000710',
                 (u'Bruker CLINPROT robot software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'GENOLINK', u'MS:1000723',
                 (u'Bruker GENOLINK software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'GenoTools', u'MS:1000724',
                 (u'Bruker GenoTools software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PolyTools', u'MS:1000727',
                 (u'Bruker PolyTools software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PROTEINEER', u'MS:1000729',
                 (u'Bruker PROTEINEER software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PureDisk', u'MS:1000735',
                 (u'BrukerPureDisk software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PROTEINEER-LC', u'MS:1000733',
                 (u'Bruker PROTEINEER-LC software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PROTEINEER spII', u'MS:1000732',
                 (u'Bruker PROTEINEER spII software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PROTEINEER fc', u'MS:1000731',
                 (u'Bruker PROTEINEER fc software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'PROTEINEER dp', u'MS:1000730',
                 (u'Bruker PROTEINEER dp software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'WARP-LC', u'MS:1000739',
                 (u'Bruker WARP-LC software.'),
                 'software',
                 [u'Bruker software', u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'TargetAnalysis', u'MS:1000738',
                 (u'Bruker TargetAnalysis software.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'HyStar', u'MS:1000817',
                 (u'Bruker software for hyphenated experiments.'),
                 'software',
                 [u'Bruker software', u'software']),
    SoftwareName(u'Anubis', u'MS:1002410',
                 (u'Anubis software for selected reaction monitoring data."'
                  u'[PSI:PI'),
                 'software',
                 [u'SRM software', u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'ATAQS', u'MS:1000925',
                 (u'Software suite used to predict, select, and optimize'
                  u'transitions as well as analyze the results of selected'
                  u'reaction monitoring runs developed and distributed by the'
                  u'Institute for Systems Biology.'),
                 'software',
                 [u'SRM software', u'software']),
    SoftwareName(u'Skyline', u'MS:1000922',
                 (u'Software used to predict, select, and optimize transitions'
                  u'as well as analyze the results of selected reaction'
                  u'monitoring runs developed and distributed by the MacCoss lab'
                  u'at the University of Washington." [https://brendanx-uw1.gs.w'
                  u'ashington.edu/labkey/wiki/home/software/Skyline/page.view?na'
                  u'me=defaul'),
                 'software',
                 [u'SRM software', u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'TIQAM', u'MS:1000923',
                 (u'Software used to predict, select, and optimize transitions'
                  u'for selected reaction monitoring experiments developed and'
                  u'distributed by the Institute for Systems Biology.'),
                 'software',
                 [u'SRM software', u'software']),
    SoftwareName(u'MaRiMba', u'MS:1000872',
                 (u'Software used to predict transitions for selected reaction'
                  u'monitoring experiments based on observed spectrum libraries'
                  u'developed and distributed by the Institute for Systems'
                  u'Biology.'),
                 'software',
                 [u'SRM software', u'software']),
    SoftwareName(u'MRMaid', u'MS:1002220',
                 (u'A web-based SRM assay design tool whose transitions are'
                  u'generated by mining the millions of identified peptide'
                  u"spectra held in the EBI's PRIDE database."),
                 'software',
                 [u'SRM software', u'software']),
    SoftwareName(u'SSRCalc', u'MS:1000874',
                 (u'Sequence Specific Retention Calculator estimates the'
                  u'retention time of peptides based on their sequence.'),
                 'software',
                 [u'peptide attribute calculation software', u'software']),
    SoftwareName(u'SILACAnalyzer', u'MS:1001831',
                 (u'Software for SILAC workflow.'),
                 'software',
                 [u'quantitation software name', u'TOPP software', u'software', u'quantification information', u'analysis software', u'data processing software']),
    SoftwareName(u'Progenesis LC-MS', u'MS:1001830',
                 (u'Software from Nonlinear Dynamics for LC-MS label-free'
                  u'workflow.'),
                 'software',
                 [u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'FindPairs', u'MS:1002063',
                 (u'Software e.g. for SILAC and 14N/15N workflow, part of the'
                  u'PeakQuant suite.'),
                 'software',
                 [u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'Microsoft Excel', u'MS:1002059',
                 (u'Microsoft Excel (can be used for spectral counting).'),
                 'software',
                 [u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'ProteoSuite', u'MS:1002124',
                 (u'ProteoSuite software for the analysis of quantitative'
                  u'proteomics data." [DOI:10.1089/omi.2012.0022, PMID:22804616'),
                 'software',
                 [u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'x-Tracker', u'MS:1002123',
                 (u'X-Tracker generic tool for quantitative proteomics.'),
                 'software',
                 [u'quantitation software name', u'software', u'quantification information']),
    SoftwareName(u'ITRAQAnalyzer', u'MS:1002129',
                 (u'Software for iTRAQ workflow. Extracts and normalizes iTRAQ'
                  u'information from an MS experiment.'),
                 'software',
                 [u'quantitation software name', u'TOPP software', u'software', u'quantification information', u'analysis software', u'data processing software']),
    SoftwareName(u'MALDI Solutions Microbial Identification', u'MS:1001878',
                 (u'Shimadzu Biotech software for data acquisition, processing,'
                  u'and analysis.'),
                 'software',
                 [u'MALDI Solutions', u'acquisition software', u'analysis software', u'data processing software', u'Shimadzu Corporation software', u'software']),
    SoftwareName(u'PRIDE Converter2', u'MS:1002335',
                 (u'Java software designed to convert one of several proteomics'
                  u'identification results formats into PRIDE XML.'),
                 'software',
                 [u'conversion software', u'data processing software', u'software']),
    SoftwareName(u'ProCon', u'MS:1002334',
                 (u'Java software designed to convert one of several proteomics'
                  u'identification results formats into mzIdentML or PRIDE XML."'
                  u'[PSI:PI'),
                 'software',
                 [u'conversion software', u'data processing software', u'software']),
    SoftwareName(u'python-psims', u'MS:1002991',
                 (u'python-psims, a library for generating mzML and mzIdentML.'),
                 'software',
                 [u'conversion software', u'data processing software', u'software']),
    SoftwareName(u'TOPP noise filter', u'MS:1002131',
                 (u'Noise filter component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP spectra filter', u'MS:1002137',
                 (u'Spectra filter component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP peak picker', u'MS:1002134',
                 (u'Peak picker component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP map aligner', u'MS:1002147',
                 (u'Map aligner component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MassTraceExtractor', u'MS:1002159',
                 (u'Annotates mass traces in centroided LC/MS maps.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MzTabExporter', u'MS:1002158',
                 (u'Exports various XML formats to an mzTab file.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDMerger', u'MS:1002155',
                 (u'Merges several protein/peptide identification files into one'
                  u'file.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP DTAExtractor', u'MS:1002154',
                 (u'Extracts spectra of an MS run file to several files in DTA'
                  u'format.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraMerger', u'MS:1002157',
                 (u'Merges spectra from an LC/MS map, either by precursor or by'
                  u'RT blocks.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDFileConverter', u'MS:1002156',
                 (u'Converts identification engine file formats.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP PrecursorMassCorrector', u'MS:1002160',
                 (u'Correct the precursor entries of tandem MS scans.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP HighResPrecursorMassCorrector', u'MS:1002161',
                 (u'Performs precursor mz correction on centroided high'
                  u'resolution data.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP AdditiveSeries', u'MS:1002162',
                 (u'Computes an additive series to quantify a peptide in a set'
                  u'of samples.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP Decharger', u'MS:1002163',
                 (u'Decharges and merges different feature charge variants of'
                  u'the same chemical entity.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP EICExtractor', u'MS:1002164',
                 (u'Quantifies signals at given positions in (raw or picked)'
                  u'LC/MS maps.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP feature finder', u'MS:1002165',
                 (u'Feature finder component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP ConsensusID', u'MS:1002188',
                 (u'Computes a consensus identification from peptide'
                  u'identifications of several identification engines.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDConflictResolver', u'MS:1002189',
                 (u'Resolves ambiguous annotations of features with peptide'
                  u'identifications.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP software adaptor', u'MS:1002180',
                 (u'Software adaptor to an external program in the TOPP'
                  u'software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpecLibSearcher', u'MS:1002187',
                 (u'Identifies peptide MS2 spectra by spectral matching with a'
                  u'searchable spectral library.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP feature linker', u'MS:1002174',
                 (u'Feature linker component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MapRTTransformer', u'MS:1002173',
                 (u'Applies retention time transformations to maps.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP ConsensusMapNormalizer', u'MS:1002172',
                 (u'Normalizes maps of one consensus XML file (after linking).'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP ProteinQuantifier', u'MS:1002171',
                 (u'Computes protein abundances from annotated feature/consensus'
                  u'maps.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP CompNovoCID', u'MS:1002179',
                 (u'Performs a peptide/protein identification with the CompNovo'
                  u'engine in collision-induced dissociation (CID) mode.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP CompNovo', u'MS:1002178',
                 (u'Performs a peptide/protein identification with the CompNovo'
                  u'engine.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP ProteinInference', u'MS:1002203',
                 (u'Infer proteins from a list of (high-confidence) peptides.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FalseDiscoveryRate', u'MS:1002204',
                 (u'Estimates the false discovery rate on peptide and protein'
                  u'level using decoy searches.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDMapper', u'MS:1002191',
                 (u'Assigns protein/peptide identifications to feature or'
                  u'consensus features.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDFilter', u'MS:1002190',
                 (u'Filters results from protein or peptide identification'
                  u'engines based on different criteria.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDRTCalibration', u'MS:1002193',
                 (u'Calibrate Retention times of peptide hits to standards.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP IDPosteriorErrorProbability', u'MS:1002192',
                 (u'Estimates posterior error probabilities using a mixture'
                  u'model.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP PrecursorIonSelector', u'MS:1002195',
                 (u'A tool for precursor ion selection based on identification'
                  u'results.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP PeptideIndexer', u'MS:1002194',
                 (u'Refreshes the protein references for all peptide hits.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OpenSwath component', u'MS:1002197',
                 (u'OpenSwath component of the TOPP software.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MRMMapper', u'MS:1002196',
                 (u'MRMMapper maps measured chromatograms (mzML) and the'
                  u'transitions used (TraML).'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'OpenXQuest', u'MS:1002673',
                 (u'Cross-Linking MS search engine.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'BaselineFilter', u'MS:1000753',
                 (u'Removes the baseline from profile spectra using a top-hat'
                  u'filter.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'DBImporter', u'MS:1000755',
                 (u'Imports data to an OpenMS database.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'DBExporter', u'MS:1000754',
                 (u'Exports data from an OpenMS database to a file.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'FileFilter', u'MS:1000757',
                 (u'Extracts or manipulates portions of data from peak, feature'
                  u'or consensus feature files.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'FileConverter', u'MS:1000756',
                 (u'Converts between different MS file formats.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'InternalCalibration', u'MS:1000759',
                 (u'Applies an internal calibration.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'FileMerger', u'MS:1000758',
                 (u'Merges several MS files into one file.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'Resampler', u'MS:1000764',
                 (u'Transforms an LC/MS map into a resampled map or a png image.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'SpectraFilter', u'MS:1000765',
                 (u'OBSOLETE Applies a filter to peak spectra.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOFCalibration', u'MS:1000766',
                 (u'Applies time of flight calibration.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MapAligner', u'MS:1000760',
                 (u'OBSOLETE Corrects retention time distortions between maps.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'MapNormalizer', u'MS:1000761',
                 (u'Normalizes peak intensities in an MS run.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'NoiseFilter', u'MS:1000762',
                 (u'OBSOLETE Removes noise from profile spectra by using'
                  u'different smoothing techniques.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'PeakPicker', u'MS:1000763',
                 (u'OBSOLETE Finds mass spectrometric peaks in profile mass'
                  u'spectra.'),
                 'software',
                 [u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteoWizard SeeMS', u'MS:1002209',
                 (u'An interactive GUI application to view and filter mass'
                  u'spectrometry data in a variety of formats.'),
                 'software',
                 [u'ProteoWizard software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteoWizard msaccess', u'MS:1002208',
                 (u'Filters, processes, and displays mass spectrometry data in a'
                  u'variety of ways.'),
                 'software',
                 [u'ProteoWizard software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteoWizard msconvert', u'MS:1002205',
                 (u'Converts, filters, and processes mass spectrometry data in'
                  u'variety of formats.'),
                 'software',
                 [u'ProteoWizard software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteoWizard chainsaw', u'MS:1002207',
                 (u'Filters and processes protein sequence databases.'),
                 'software',
                 [u'ProteoWizard software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'ProteoWizard idconvert', u'MS:1002206',
                 (u'Converts, filters, and processes identifications from'
                  u'shotgun proteomics experiments.'),
                 'software',
                 [u'ProteoWizard software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'mzidLib:Omssa2Mzid', u'MS:1002238',
                 (u'A converter for OMSSA OMX to mzIdentML.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:Tandem2Mzid', u'MS:1002239',
                 (u'A converter for Tandem XML to mzIdentML.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:Csv2Mzid', u'MS:1002240',
                 (u'A converter for CSV files (following OMSSA CSV style) to'
                  u'mzIdentML.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:Mzidentml2Csv', u'MS:1002245',
                 (u'A tool for converting mzIdentML files to CSV format.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:FalseDiscoveryRate', u'MS:1002244',
                 (u'A routine for calculating local FDR, q-value and FDRScore'
                  u'for mzIdentML files, based on a decoy search.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:InsertMetaDataFromFasta', u'MS:1002247',
                 (u'A tool for adding additional meta data from a FASTA file to'
                  u'DBSequence entries (sequence and description) in mzIdentML'
                  u'files.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:CombineSearchEngines', u'MS:1002246',
                 (u'A tool for combining results analysed in parallel in two or'
                  u'three search engines into a single mzIdentML file.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:ProteoGrouper', u'MS:1002241',
                 (u'A generic and parameterizable protein inference algorithm'
                  u'for mzIdentML files.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:Perform emPAI on mzid', u'MS:1002243',
                 (u'A routine for adding emPAI quantitative values to an'
                  u'mzIdentML file.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'mzidLib:Thresholder', u'MS:1002242',
                 (u'A routine for keeping only identifications passing a given'
                  u'threshold or setting passThreshold to true or false for'
                  u'SpectrumIdentificationItem or ProteinDetectionHypothesis in'
                  u'mzIdentML files.'),
                 'software',
                 [u'mzidLib', u'analysis software', u'software']),
    SoftwareName(u'lipidomics analysis software', u'MS:1002964',
                 (u'Lipidomics analysis software.'),
                 'software',
                 [u'small molecule analysis software', u'analysis software', u'software']),
    SoftwareName(u'Progenesis QI', u'MS:1002879',
                 (u'Metabolomics analysis software for LC-MS data from Nonlinear'
                  u'Dynamics.'),
                 'software',
                 [u'small molecule analysis software', u'analysis software', u'software']),
    SoftwareName(u'MyCompoundID', u'MS:1002881',
                 (u'Metabolite identification tool MyCompoundID." [PSI:PI'),
                 'software',
                 [u'small molecule analysis software', u'analysis software', u'software']),
    SoftwareName(u'Compound Discoverer', u'MS:1002880',
                 (u'Metabolomics analysis software from Thermo Fisher'
                  u'Scientific.'),
                 'software',
                 [u'small molecule analysis software', u'analysis software', u'software']),
    SoftwareName(u'ASAPRatio', u'MS:1002574',
                 (u'A program in the TPP that calculates PSM, peptide, and'
                  u'protein-level abundances based on 2-channel isotope-labelled'
                  u'data such as ICAT, SILAC, etc.'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'PTMProphet', u'MS:1002292',
                 (u'A program in the TPP that calculates PTM localization'
                  u'probabilities by re-analyzing the peaks that are available'
                  u'to distinguish between possible modification sites.'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'XPRESS', u'MS:1002290',
                 (u'A program in the TPP that calculates PSM-level abundances'
                  u'based on 2-channel isotope-labelled data such as ICAT,'
                  u'SILAC, etc.'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'Libra', u'MS:1002291',
                 (u'A program in the TPP that calculates PSM, peptide, and'
                  u'protein-level abundances based on N-channel isobaric label'
                  u'peptide data such as iTRAQ, TMT, etc.'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'PeptideProphet', u'MS:1002287',
                 (u'A program in the TPP that calculates PSM probabilities for'
                  u'MS2 proteomics data searched with any of the supported'
                  u'sequence or spectral library search engines via the pepXML'
                  u'format." [PMID:12403597'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'ProteinProphet', u'MS:1002289',
                 (u'A program in the TPP that calculates protein-level'
                  u'probabilities based on input PSM or peptide-level'
                  u'probabilities from PeptideProphet or iProphet. The output is'
                  u'written in the protXML format.'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'iProphet', u'MS:1002288',
                 (u'A program in the TPP that calculates distinct peptide'
                  u'probabilities based on several lines of corroborating'
                  u'evidence including search results from multiple search'
                  u'engines via the pepXML format.'),
                 'software',
                 [u'Trans-Proteomic Pipeline software', u'analysis software', u'software']),
    SoftwareName(u'TOPP NoiseFilterSGolay', u'MS:1002133',
                 (u'Removes noise from profile spectra by using a Savitzky-Golay'
                  u'smoothing.'),
                 'software',
                 [u'TOPP noise filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP NoiseFilterGaussian', u'MS:1002132',
                 (u'Removes noise from profile spectra by using a gaussian'
                  u'smoothing.'),
                 'software',
                 [u'TOPP noise filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterMarkerMower', u'MS:1002139',
                 (u'Applies a filter to peak spectra for marked peaks.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterBernNorm', u'MS:1002138',
                 (u'Applies a Bern et al normalization to peak spectra."'
                  u'[PMID:15262780'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterWindowMower', u'MS:1002146',
                 (u'Applies a filter of the largest peaks in a sliding window'
                  u'over a peak spectrum.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterSqrtMower', u'MS:1002144',
                 (u'Applies a filter to peak spectra after intensity scaling to'
                  u'the square root.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterThresholdMower', u'MS:1002145',
                 (u'Applies a filter of peaks below a given threshold to peak'
                  u'spectra.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterParentPeakMower', u'MS:1002142',
                 (u'Filters putative unfragmented precursor ions from tandem'
                  u'spectra.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterScaler', u'MS:1002143',
                 (u'Applies a filter to peak spectra after intensity scaling'
                  u'according to rank.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterNLargest', u'MS:1002140',
                 (u'Retains the n largest peaks of a peak spectra.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP SpectraFilterNormalizer', u'MS:1002141',
                 (u'Applies a TIC/maximal intensity normalization to peak'
                  u'spectra.'),
                 'software',
                 [u'TOPP spectra filter', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP PeakPickerWavelet', u'MS:1002136',
                 (u'Finds mass spectrometric peaks with a wavelet algorithm in'
                  u'low-resoluted profile mass spectra.'),
                 'software',
                 [u'TOPP peak picker', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP PeakPickerHiRes', u'MS:1002135',
                 (u'Finds mass spectrometric peaks in high-resoluted profile'
                  u'mass spectra.'),
                 'software',
                 [u'TOPP peak picker', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MapAlignerIdentification', u'MS:1002148',
                 (u'Corrects retention time distortions between maps based on'
                  u'common peptide identifications.'),
                 'software',
                 [u'TOPP map aligner', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MapAlignerPoseClustering', u'MS:1002149',
                 (u'Corrects retention time distortions between maps using a'
                  u'pose clustering approach.'),
                 'software',
                 [u'TOPP map aligner', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MapAlignerSpectrum', u'MS:1002150',
                 (u'Corrects retention time distortions between maps by spectrum'
                  u'alignment.'),
                 'software',
                 [u'TOPP map aligner', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureFinderCentroided', u'MS:1002166',
                 (u'Detects two-dimensional features in centroided LC-MS data.'),
                 'software',
                 [u'TOPP feature finder', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureFinderRaw', u'MS:1002167',
                 (u'Detects two-dimensional features in uncentroided LC-MS data.'),
                 'software',
                 [u'TOPP feature finder', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureFinderIsotopeWavelet', u'MS:1002168',
                 (u'Detects two-dimensional features in uncentroided LC-MS data'
                  u'with a wavelet algorithm.'),
                 'software',
                 [u'TOPP feature finder', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureFinderMetabo', u'MS:1002169',
                 (u'Detects two-dimensional features in centroided LC-MS data of'
                  u'metabolites.'),
                 'software',
                 [u'TOPP feature finder', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureFinderMRM', u'MS:1002170',
                 (u'Quantifies features LC-MS/MS MRM data.'),
                 'software',
                 [u'TOPP feature finder', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MascotAdapter', u'MS:1002182',
                 (u'Identifies MS2 spectra using the external program Mascot.'),
                 'software',
                 [u'TOPP software adaptor', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP MascotAdapterOnline', u'MS:1002183',
                 (u'Identifies MS2 spectra using the online version of the'
                  u'external program Mascot.'),
                 'software',
                 [u'TOPP software adaptor', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP InspectAdapter', u'MS:1002181',
                 (u'Identifies MS2 spectra using the external program Inspect.'),
                 'software',
                 [u'TOPP software adaptor', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP XTandemAdapter', u'MS:1002186',
                 (u'Identifies MS2 spectra using the external program XTandem.'),
                 'software',
                 [u'TOPP software adaptor', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OMSSAAdapter', u'MS:1002184',
                 (u'Identifies MS2 spectra using the external program OMSSA.'),
                 'software',
                 [u'TOPP software adaptor', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP PepNovoAdapter', u'MS:1002185',
                 (u'Identifies MS2 spectra using the external program PepNovo.'),
                 'software',
                 [u'TOPP software adaptor', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureLinkerUnlabeledQT', u'MS:1002177',
                 (u'Groups corresponding features from multiple maps using a'
                  u'quality threshold clustering approach.'),
                 'software',
                 [u'TOPP feature linker', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureLinkerUnlabeled', u'MS:1002176',
                 (u'Groups corresponding features from multiple maps.'),
                 'software',
                 [u'TOPP feature linker', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP FeatureLinkerLabeled', u'MS:1002175',
                 (u'Groups corresponding isotope-labeled features in a feature'
                  u'map.'),
                 'software',
                 [u'TOPP feature linker', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OpenSwathFeatureXMLToTSV', u'MS:1002201',
                 (u'Converts a featureXML to a mProphet tsv (tab separated'
                  u'values).'),
                 'software',
                 [u'TOPP OpenSwath component', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OpenSwathDecoyGenerator', u'MS:1002200',
                 (u'Generates decoys according to different models for a'
                  u'specific TraML.'),
                 'software',
                 [u'TOPP OpenSwath component', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OpenSwathRTNormalizer', u'MS:1002202',
                 (u'Generates a transformation file for retention time space'
                  u'into normalized space.'),
                 'software',
                 [u'TOPP OpenSwath component', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OpenSwathChromatogramExtractor', u'MS:1002199',
                 (u'Extract chromatograms (XIC) from a MS2 map file.'),
                 'software',
                 [u'TOPP OpenSwath component', u'TOPP software', u'analysis software', u'data processing software', u'software']),
    SoftwareName(u'TOPP OpenSwathAnalyzer', u'MS:1002198',
                 (u'Picks peaks and finds features in an SRM experiment.'),
                 'software',
                 [u'TOPP OpenSwath component', u'TOPP software', u'analysis software', u'data processing software', u'software']),
])
# [[[end]]]


software_names_by_name = {c.name: c for c in software_names}


def software_name(name):
    try:
        return software_names_by_name[name]
    except KeyError:
        return SoftwareName(name, name, name, name, [name])


__all__ = [
    "SoftwareName", "Software", "software_names",
    "software_name"
]



from cpython.list cimport PyList_GetItem, PyList_Size
from cpython.tuple cimport PyTuple_GetItem

from ms_deisotope._c.peak_dependency_network.intervals cimport (SpanningMixin, Interval)
from ms_deisotope._c.scoring cimport IsotopicFitRecord


cdef class FitNode(SpanningMixin):

    cdef:
        public size_t index
        public IsotopicFitRecord fit
        public set edges
        public set overlap_edges
        public set peak_indices
        public long _hash
        public double score

    cpdef bint isdisjoint(self, FitNode other)
    cpdef visit(self, FitNode other)

    cdef void _init_fields(self, list experimental)

    @staticmethod
    cdef FitNode _create(IsotopicFitRecord fit, size_t index)


cdef class ConnectedSubgraph(object):
    cdef:
        public tuple nodes
        public bint maximize

    cdef void _init_nodes(self, list fits)
    cdef void populate_edges(self)
    cpdef find_heaviest_path(self, method=*)



import os
from .mzml import MzMLLoader
from .mzxml import MzXMLLoader
from .mgf import MGFLoader
from . import _compression

guessers = []
reader_types = [MzMLLoader, MzXMLLoader, MGFLoader]


def register_type_guesser(reader_guesser):
    guessers.append(reader_guesser)
    return reader_guesser


@register_type_guesser
def guess_type_from_path(file_path):
    if hasattr(file_path, 'name'):
        file_path = file_path.name
    ext = os.path.splitext(file_path)[1].lower()
    if ext == '.mzml':
        return MzMLLoader
    elif ext == '.mzxml':
        return MzXMLLoader
    elif ext == '.mgf':
        return MGFLoader
    else:
        raise ValueError("Cannot determine ScanLoader type from file path")


@register_type_guesser
def guess_type_from_file_sniffing(file_path):
    is_random_access_file = is_random_access(file_path)
    if is_random_access_file:
        handle = file_path
        header = handle.read(1000)
        handle.seek(0)
    else:
        with open(file_path, 'rb') as handle:
            header = handle.read(1000)

    if _compression.starts_with_gz_magic(header):
        if is_random_access_file:
            handle = _compression.GzipFile(fileobj=file_path, mode='rb')
            header = handle.read(1000)
            handle.seek(0)
        else:
            with _compression.GzipFile(file_path, mode='rb') as handle:
                header = handle.read(1000)

    if b"mzML" in header:
        return MzMLLoader
    elif b"mzXML" in header:
        return MzXMLLoader
    elif b"BEGIN IONS" in header:
        return MGFLoader
    else:
        raise ValueError("Cannot determine ScanLoader type from header")


def guess_type(file_path):
    for guesser in guessers:
        try:
            reader_type = guesser(file_path)
            return reader_type
        except (ValueError, IOError, ImportError, TypeError, AttributeError):
            continue
    raise ValueError("Cannot determine ScanLoader type")


def MSFileLoader(file_path, *args, **kwargs):
    """Factory function to create an object that reads scans from
    any supported data file format. Provides both iterative and
    random access.
    """
    reader_type = guess_type(file_path)
    try:
        is_file = os.path.isfile(file_path)
    except TypeError:
        is_file = False
    if is_file:
        is_gz_compressed = _compression.test_gzipped(file_path)
        if is_gz_compressed:
            fobj = _compression.get_opener(file_path)
        else:
            fobj = file_path
    else:
        fobj = file_path
    return reader_type(fobj, *args, **kwargs)


def is_random_access(fp):
    return hasattr(fp, 'read') and hasattr(fp, 'seek')


try:
    from .thermo_raw import (
        ThermoRawLoader, infer_reader as _check_is_thermo_raw,
        register_dll as register_thermo_dll)

    reader_types.append(ThermoRawLoader)
    register_type_guesser(_check_is_thermo_raw)

except ImportError:  # pragma: no cover
    def register_thermo_dll(*args, **kwargs):
        pass

try:
    from .thermo_raw_net import (
        ThermoRawLoader as ThermoRawNetLoader, infer_reader as _check_is_thermo_raw_net,
        register_dll as register_thermo_net_dll)
    reader_types.append(ThermoRawNetLoader)
    register_type_guesser(_check_is_thermo_raw_net)

except ImportError:  # pragma: no cover
    def register_thermo_net_dll(*args, **kwargs):
        pass

try:
    from .agilent_d import (
        AgilentDLoader,
        infer_reader as _check_is_agilent_d,
        register_dll_dir as register_agilent_dll_dir)

    reader_types.append(AgilentDLoader)
    register_type_guesser(_check_is_agilent_d)
except ImportError:  # pragma: no cover
    def register_agilent_dll_dir(*args, **kwargs):
        pass



from cpython.list cimport PyList_GetItem, PyList_Size
from cpython.tuple cimport PyTuple_GetItem

from collections import deque


cdef class SpanningMixin(object):
    """Provides methods for checking whether an entity
    which has a defined start and end point over a single
    dimension contains or overlaps with another entity in
    that same dimension.
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __contains__(self, i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : Number
            The point to be tested

        Returns
        -------
        bool
        """
        return self._contains(i)

    cdef bint _contains(self, double i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : double
            The point to be tested

        Returns
        -------
        bool
        """
        return self.start <= i <= self.end

    cpdef bint contains(self, double i):
        """Tests for point inclusion, `start <= i <= end`

        Parameters
        ----------
        i : double
            The point to be tested

        Returns
        -------
        bool
        """
        return self._contains(i)

    cpdef bint overlaps(self, SpanningMixin interval):
        """Tests whether another spanning entity with
        a defined start and end point overlaps with
        this spanning entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        cdef bint cond
        cond = ((self.start <= interval.start and self.end >= interval.start) or (
            self.start >= interval.start and self.end <= interval.end) or (
            self.start >= interval.start and self.end >= interval.end and self.start <= interval.end) or (
            self.start <= interval.start and self.end >= interval.start) or (
            self.start <= interval.end and self.end >= interval.end))
        return cond

    cpdef double overlap_size(self, SpanningMixin interval):
        if self.start <= interval.start <= self.end and self.start <= interval.end <= self.end:
            return interval.end - interval.start
        elif interval.start <= self.start <= interval.end and interval.start <= self.end <= interval.end:
            return self.end - self.start
        if self.end >= interval.start:
            return self.end - interval.start
        elif self.start >= interval.end:
            return self.start - interval.end

    cpdef bint is_contained_in_interval(self, SpanningMixin interval):
        """Tests whether this spanning entity is
        completely contained inside the other entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        return self.start >= interval.start and self.end <= interval.end

    cpdef bint contains_interval(self, SpanningMixin interval):
        """Tests whether the other spanning entity is
        completely contained inside this entity

        Parameters
        ----------
        interval : SpanningMixin

        Returns
        -------
        bool
        """
        return self.start <= interval.start and self.end >= interval.end


cdef class Interval(SpanningMixin):
    def __init__(self, start, end, members=None, **kwargs):
        if members is None:
            members = []
        self.start = start
        self.end = end
        self.members = members
        self.data = kwargs

    def __repr__(self):
        return "%s(start=%r, end=%r, data=%r)" % (self.__class__.__name__, self.start, self.end, self.data)

    def __getitem__(self, k):
        return self.members[k]

    def __len__(self):
        return len(self.members)

    def __iter__(self):
        return iter(self.members)


cpdef list intervals_containing_point(list intervals, double x):
    cdef:
        size_t i
        SpanningMixin interval
        list result
    result = []
    for i in range(PyList_Size(intervals)):
        interval = <SpanningMixin>PyList_GetItem(intervals, i)
        if interval.start <= x <= interval.end:
            result.append(interval)
    return result


cdef double INF = float('inf')


cdef class IntervalTreeNode(object):

    def __init__(self, double center, IntervalTreeNode left, list contained, IntervalTreeNode right,
                 int level=0, IntervalTreeNode parent=None):
        cdef:
            double start, end, i_start, i_end
            size_t i

        self.center = center
        self.left = left
        self.contained = contained
        self.right = right
        self.level = level
        self.parent = parent

        start = INF
        end = -INF
        i = 0

        for interval in self.contained:
            i += 1
            i_start = interval.start
            if i_start < start:
                start = i_start
            i_end = interval.end
            if i_end > end:
                end = i_end

        if i > 0:
            self.start = start
            self.end = end
        else:
            self.start = self.end = center

    cpdef list contains_point(self, double x):
        cdef:
            list inner
        inner = intervals_containing_point(self.contained, x)
        if self.left is not None and self.left.start <= x <= self.left.end:
            inner.extend(self.left.contains_point(x))
            return inner
        if self.right is not None and self.right.start <= x <= self.right.end:
            inner.extend(self.right.contains_point(x))
            return inner
        else:
            return inner

    cpdef list _overlaps_interval(self, double start, double end):
        cdef:
            SpanningMixin interv
            size_t i

        result = []
        for i in range(PyList_Size(self.contained)):
            interv = <SpanningMixin>PyList_GetItem(self.contained, i)
            if ((start < interv.start and interv.end <= end) or
                    (interv.start <= start and interv.end >= end)):
                result.append(interv)
        return result

    cpdef list overlaps(self, double start, double end):
        """Returns the list of all contained intervals which
        overlap with the interval described by `start` and `end`

        Parameters
        ----------
        start : float
            The start of the query interval
        end : float
            The end of the query interval

        Returns
        -------
        list
            A list of all objects which overlap the argument interval
            from all spanning nodes' `contained` list.
        """
        cdef list result = []
        if start > self.end:
            return result
        if end < self.start:
            return result
        elif start <= self.start:
            if end < self.start:
                return result
            else:
                if self.left is not None:
                    result.extend(self.left.overlaps(start, end))
                result.extend(self._overlaps_interval(start, end))
                if self.right is not None and end >= self.right.start:
                    result.extend(self.right.overlaps(start, end))
        elif start > self.start:
            if self.left is not None and self.left.end >= start:
                result.extend(self.left.overlaps(start, end))
            result.extend(self._overlaps_interval(start, end))
            if self.right is not None and end >= self.right.start:
                result.extend(self.right.overlaps(start, end))
        elif end > self.start:
            if self.left is not None:
                result.extend(self.left.overlaps(start, end))
            result.extend(self._overlaps_interval(start, end))
            if self.right is not None and end >= self.right.start:
                result.extend(self.right.overlaps(start, end))
        return result

    def __repr__(self):
        return "IntervalTreeNode(level=%d, center=%0.4f, start=%0.4f, end=%0.4f)" % (
            self.level, self.center, self.start, self.end)

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))

    cpdef bint _eq(self, IntervalTreeNode other):
        if other is None:
            return False
        else:
            result = self.contained == other.contained
            if result:
                result = self.left == other.left
                if result:
                    result = self.right == other.right
            return result

    def __hash__(self):
        return hash((self.start, self.center, self.right, self.level))

    def __diagnostic_eq__(self, other):
        if other is None:
            return False
        else:
            result = self.contained == other.contained
            if result:
                try:
                    result = self.left.__diagnostic_eq__(other.left)
                except AttributeError:
                    result = self.left == other.left
                if result:
                    try:
                        result = self.right.__diagnostic_eq__(other.right)
                    except AttributeError:
                        result = self.right == other.right
                    if not result:
                        print self.right, "r!=", other.right
                else:
                    print self.left, "l!=", other.left
            else:
                print self, "!=", other
            return result

    @classmethod
    def build(cls, intervals):
        cdef:
            IntervalTreeNode root, parent, node, up
            list stack, members, left, right, contained, intervals_
            tuple bunch
            str side
            double center, point, accumulator
            SpanningMixin i
            size_t j, n
        stack = []
        root = cls(0, None, [], None, -1)

        intervals_ = list(intervals)
        if not intervals_:
            return root
        stack.append((root, intervals_, "left"))

        while stack:
            # parent, members, side = stack.pop()
            bunch = <tuple>stack.pop()
            parent = <IntervalTreeNode>PyTuple_GetItem(bunch, 0)
            members = <list>PyTuple_GetItem(bunch, 1)
            side = <str>PyTuple_GetItem(bunch, 2)
            n = PyList_Size(members)
            if n > 0:
                accumulator = 0.
                j = 0
                for j in range(n):
                    i = <SpanningMixin>PyList_GetItem(members, j)
                    accumulator += (i.start + i.end) / 2.
                if j > 0:
                    center = accumulator / (j + 1)
                else:
                    center = 0.0

                left = []
                right = []
                contained = []

                if n > 20:
                    j = 0
                    for j in range(n):
                        i = <SpanningMixin>PyList_GetItem(members, j)
                        if abs(i.start - center) < 1e-6 and abs(i.end - center) < 1e-6:
                            contained.append(i)
                        elif center > i.end:
                            left.append(i)
                        elif center < i.start:
                            right.append(i)
                        else:
                            contained.append(i)
                else:
                    contained = list(members)
                node = cls(center, left=None, contained=contained, right=None, level=parent.level + 1, parent=parent)
                if side == 'left':
                    parent.left = node
                    parent.start = min(node.start, parent.start)
                    up = parent.parent
                    while up is not None:
                        if up.start > node.start:
                            up.start = node.start
                        up = up.parent
                elif side == 'right':
                    parent.right = node
                    parent.end = max(node.end, parent.end)
                    up = parent.parent
                    while up is not None:
                        if up.end < node.end:
                            up.end = node.end
                        up = up.parent
                else:
                    raise ValueError(side)
                stack.append((node, left, "left"))
                stack.append((node, right, "right"))
        return root.left

    def _update_bounds(self):
        changed = False
        if self.left is not None and self.left.start < self.start:
            changed = True
            self.start = self.left.start
        if self.right is not None and self.right.end > self.end:
            changed = True
            self.end = self.right.end
        if self.parent is not None:
            self.parent._update_bounds()
        return changed

    def insert(self, interval):
        insert_in_self = False
        if self.node_contains(interval.start):
            if self.left is not None and self.left.node_contains(interval.end):
                return self.left.insert(interval)
            elif (self.right is not None and self.right.node_contains(interval.end) and
                  self.right.node_contains(interval.start)):
                return self.right.insert(interval)
            else:
                insert_in_self = True
        elif self.node_contains(interval.end):
            if self.right is not None and self.right.node_contains(interval.start):
                return self.right.insert(interval)
            else:
                insert_in_self = True
        if not insert_in_self and self.parent is None:
            insert_in_self = True
        if insert_in_self:
            self.contained.append(interval)
            changed = False
            if interval.start < self.start:
                self.start = interval.start
                changed = True
            if interval.end > self.end:
                self.end = interval.end
                changed = True
            if changed:
                self._update_bounds()

    def flatten(self):
        # perform an infix traversal of the tree and collect
        # the elements
        items = []
        stack = deque([])
        current = self
        done = False
        while not done:
            if current is not None:
                stack.append(current)
                current = current.left
            else:
                if stack:
                    current = stack.pop()
                    items.extend(current.contained)
                    current = current.right
                else:
                    done = True
        return items

    def balance(self):
        items = self.flatten()
        tree = self.build(items)
        self.left = tree.left
        self.right = tree.right
        self.contained = tree.contained
        self.start = tree.start
        self.end = tree.end
        self.center = tree.center
        self.left.parent = self
        self.right.parent = self
        self._update_bounds()

    def __iter__(self):
        flat = self.flatten()
        return iter(flat)


from itertools import cycle
try:
    from matplotlib import pyplot as plt
    from matplotlib import patches as mpatches, colors as mcolors
except (RuntimeError, ImportError):
    pass
import numpy as np


from .profile_transform import interpolate, smooth_leveled


Ellipse = mpatches.Ellipse
FancyBboxPatch = mpatches.FancyBboxPatch
Rectangle = mpatches.Rectangle


def draw_features(features, ax=None, alpha=0.65, width=2e-5, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(1)

    ellipses = []
    kwargs.setdefault("lw", 0.05)
    lw = kwargs.get("linewidth", kwargs.get("lw"))
    for feat in features:
        if feat is None:
            continue
        center = (feat.end_time + feat.start_time) / 2.
        height = feat.end_time - feat.start_time

        center_mz = feat.mz
        mz_width = center_mz * width

        ellipses.append(
            FancyBboxPatch((feat.mz - mz_width / 4., center - height / 2.), width=mz_width / 2., height=height,
                           boxstyle=mpatches.BoxStyle.Round(pad=mz_width / 2.)))

    for ell in ellipses:
        ell.set_alpha(alpha)
        ell.set_facecolor("blue")
        ell.set_edgecolor("blue")
        ell.set_linewidth(lw)
        ax.add_artist(ell)

    ax.set_xlim(
        min(features, key=lambda x: x.mz if x is not None else float('inf')).mz - 1,
        max(features, key=lambda x: x.mz if x is not None else -float('inf')).mz + 1)
    ax.set_ylim(
        min(features, key=lambda x: x.start_time if x is not None else float('inf')).start_time - 1,
        max(features, key=lambda x: x.end_time if x is not None else -float('inf')).end_time + 1)
    return ax


def draw_feature_sets(feature_sets, ax=None, alpha=0.65, width=2e-5, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(1)

    kwargs.setdefault("lw", 0.05)
    lw = kwargs.get("linewidth", kwargs.get("lw"))

    features = []
    ellipse_sets = []
    for feature_set in feature_sets:
        ellipses = []
        for feat in feature_set:
            if feat is None:
                continue
            center = (feat.end_time + feat.start_time) / 2.
            height = feat.end_time - feat.start_time

            center_mz = feat.mz
            mz_width = center_mz * width

            ellipses.append(
                FancyBboxPatch((feat.mz - mz_width / 4., center - height / 2.),
                               width=mz_width / 2., height=height,
                               boxstyle=mpatches.BoxStyle.Round(pad=mz_width / 2.)))
            features.append(feat)
        ellipse_sets.append(ellipses)

    for ellipses in ellipse_sets:
        color = np.random.rand(3)
        for ell in ellipses:
            ell.set_alpha(alpha)
            ell.set_facecolor(color)
            ell.set_edgecolor(color)
            ell.set_linewidth(lw)
            ax.add_artist(ell)

    ax.set_xlim(
        min(features, key=lambda x: x.mz if x is not None else float('inf')).mz - 1,
        max(features, key=lambda x: x.mz if x is not None else -float('inf')).mz + 1)
    ax.set_ylim(
        min(features, key=lambda x: x.start_time if x is not None else float('inf')).start_time - 1,
        max(features, key=lambda x: x.end_time if x is not None else -float('inf')).end_time + 1)
    return ax


_nice_color_cycle = ([
    'blue', 'red', 'green', 'pink', 'orange', "grey", "purple",
    "seagreen", "darkgoldenrod", "darkcyan", "skyblue",
    "maroon", "darkgreen", "slategrey", "darkslateblue"
])


nice_color_cycle = cycle(_nice_color_cycle)


def random_colorizer(profile, *args, **kwargs):
    return mcolors.rgb2hex(np.random.rand(3))


def nice_colorizer(profile, *args, **kwargs):
    return next(nice_color_cycle)


def labeler(profile, *args, **kwargs):
    label = "%0.4f" % profile.mz
    if hasattr(profile, 'charge'):
        label += ", %d" % profile.charge
    return label


def draw_profiles(profiles, ax=None, smooth=False, interp=False, label_font_size=10,
                  axis_label_font_size=20, axis_font_size=16, label=True,
                  colorizer=random_colorizer, label_function=labeler):
    if ax is None:
        fig, ax = plt.subplots(1)
    minimum_ident_time = float("inf")
    maximum_ident_time = 0
    maximum_intensity = 0

    _label_apexes = label and (label_function is not None)

    for profile in profiles:
        if profile is None:
            continue
        rt, heights = profile.as_arrays()

        if smooth:
            heights = smooth_leveled(rt, heights, smooth)
        if interp:
            rt, heights = interpolate(rt, heights)

        maximum_ident_time = max(max(rt), maximum_ident_time)
        minimum_ident_time = min(min(rt), minimum_ident_time)

        label = label_function(profile) if label_function is not None else None

        maximum_intensity = max(max(heights), maximum_intensity)
        color = colorizer(profile)
        ax.scatter(rt, heights, color=color)
        ax.fill_between(
            rt,
            heights,
            alpha=0.25,
            color=color,
            label=label
        )
        apex = max(heights)
        apex_ind = np.argmax(heights)
        rt_apex = rt[apex_ind]
        if _label_apexes:
            ax.text(rt_apex, apex + min(apex * (
                .1), 1200), label,
                ha='center', fontsize=label_font_size, alpha=0.75)
    ax.set_xlim(minimum_ident_time - 0.02,
                maximum_ident_time + 0.02)
    ax.set_ylim(0, maximum_intensity * 1.1)
    ax.legend(bbox_to_anchor=(1.7, 1.), ncol=2, fontsize=10)
    ax.axes.spines['right'].set_visible(False)
    ax.axes.spines['top'].set_visible(False)
    ax.yaxis.tick_left()
    ax.xaxis.tick_bottom()
    ax.set_xlabel("Retention Time", fontsize=axis_label_font_size)
    ax.set_ylabel("Relative Abundance", fontsize=axis_label_font_size)
    [t.set(fontsize=axis_font_size) for t in ax.get_xticklabels()]
    [t.set(fontsize=axis_font_size) for t in ax.get_yticklabels()]
    return ax



from collections import OrderedDict
import bisect
from .common import (
    ScanBunch, ScanDataSource, RandomAccessScanSource,
    Scan, ProcessedScan, WrappedScan)

try:
    range = xrange
except NameError:
    pass


class MemoryScanInterface(ScanDataSource):

    @classmethod
    def make_scan(cls, arrays=None, ms_level=None, id=None, index=None, scan_time=None,
                  is_profile=False, polarity=1, precursor_information=None, activation=None,
                  isolation_window=None, annotations=None, acquisition_information=None,
                  instrument_configuration=None, peak_set=None, deconvoluted_peak_set=None,
                  **kwargs):
        '''
        Parameters
        ----------
        arrays: :class:`RawDataArrays`
            A pair of :class:`numpy.ndarray` objects corresponding to the raw m/z and intensity data points
        ms_level: int
            The degree of fragmentation performed. 1 corresponds to a MS1 or "Survey" scan, 2 corresponds
            to MS/MS, and so on. If :attr:`ms_level` > 1, the scan is considered a "tandem scan" or "MS^n" scan
        id: str
            The unique identifier for this scan as given by the source
        index: int
            The integer number indicating how many scans were acquired prior to this scan.
        scan_time: float
            The time the scan was acquired during data acquisition. The unit of time will always be minutes.
        is_profile: bool
            Whether this scan's raw data points corresponds to a profile scan or whether the raw data was
            pre-centroided.
        polarity: int
            If the scan was acquired in positive mode, the value ``+1``.  If the scan was acquired in negative
            mode, the value ``-1``. May be used to indicating how to calibrate charge state determination methods.
        precursor_information: :class:`PrecursorInformation` or None
            Descriptive metadata for the ion which was chosen for fragmentation, and a reference to
            the precursor scan
        activation: :class:`.ActivationInformation` or None
            If this scan is an MS^n scan, this attribute will contain information about the process
            used to produce it from its parent ion.
        instrument_configuration: :class:`~.InstrumentInformation`
            The instrument configuration used to acquire this scan.
        acquisition_information: :class:`.ScanAcquisitionInformation` or None
            Describes the type of event that produced this scan, as well as the scanning method
            used.
        isolation_window: :class:`.IsolationWindow` or None
            Describes the range of m/z that were isolated from a parent scan to create this scan
        annotations: dict
            A set of key-value pairs describing the scan not part of the standard interface
        peak_set : :class:`ms_peak_picker.PeakSet` or None
            Picked peaks and (possibly) associated raw data points as produced by :meth:`pick_peaks`.
            Will be `None` if peak picking has not been done.
        deconvoluted_peak_set : :class:`ms_deisotope.DeconvolutedPeakSet` or None
            Deconvoluted peaks resulting from charge state deconvolution and deisotoping. Will
            be `None` if deconvolution has not been done.
        '''
        if annotations is None:
            annotations = {}
        if arrays is None:
            arrays = [[], []]
        annotations.update(kwargs)
        scan = WrappedScan({}, cls(), arrays, None, annotations=annotations)
        scan.arrays = arrays
        scan.ms_level = ms_level
        scan.id = str(id)
        scan.title = str(id)
        scan.index = index or -1
        scan.scan_time = scan_time or -1
        scan.is_profile = bool(is_profile)
        scan.polarity = int(polarity)
        scan.activation = activation
        scan.isolation_window = isolation_window
        scan.acquisition_information = acquisition_information
        scan.instrument_configuration = instrument_configuration
        scan.peak_set = peak_set
        scan.deconvoluted_peak_set = deconvoluted_peak_set
        return scan


    def _scan_arrays(self, scan):
        """Returns raw data arrays for m/z and intensity

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        mz: np.array
            An array of m/z values for this scan
        intensity: np.array
            An array of intensity values for this scan
        """
        return None

    def _precursor_information(self, scan):
        """Returns information about the precursor ion,
        if any, that this scan was derived form.

        Returns `None` if this scan has no precursor ion

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        PrecursorInformation
        """
        return None

    def _scan_title(self, scan):
        """Returns a verbose name for this scan, if one
        were stored in the file. Usually includes both the
        scan's id string, as well as information about the
        original file and format.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        return None

    def _scan_id(self, scan):
        """Returns the scan's id string, a unique
        identifier for this scan in the context of
        the data file it is recordered in

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        try:
            return self._scan_index_map[scan].id
        except AttributeError:
            return None

    def _scan_index(self, scan):
        """Returns the base 0 offset from the start
        of the data file in number of scans to reach
        this scan.

        If the original format does not natively include
        an index value, this value may be computed from
        the byte offset index.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        return None

    def _ms_level(self, scan):
        """Returns the degree of exponential fragmentation
        used to produce this scan.

        1 refers to a survey scan of unfragmented ions, 2
        refers to a tandem scan derived from an ms level 1
        ion, and so on.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        try:
            return self._scan_index_map[scan].ms_level
        except AttributeError:
            return None

    def _scan_time(self, scan):
        """Returns the time in minutes from the start of data
        acquisition to when this scan was acquired.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        float
        """
        return None

    def _is_profile(self, scan):
        """Returns whether the scan contains profile data (`True`)
        or centroided data (`False`).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        bool
        """
        return None

    def _polarity(self, scan):
        """Returns whether this scan was acquired in positive mode (+1)
        or negative mode (-1).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        return None

    def _activation(self, scan):
        """Returns information about the activation method used to
        produce this scan, if any.

        Returns `None` for MS1 scans

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        ActivationInformation
        """
        return None


make_scan = MemoryScanInterface.make_scan


class MemoryScanLoader(MemoryScanInterface, RandomAccessScanSource):

    def __init__(self, scans, binds=True, **kwargs):
        self._scans = sorted(scans, key=lambda x: x.scan_time)
        self._binds = binds
        self._producer = None
        self._scan_id_map = dict()
        self._scan_index_map = OrderedDict()
        self._index = OrderedDict()
        self._build_indices()
        self.make_iterator()

    @property
    def _scan_cache(self):
        return self._scan_id_map

    @classmethod
    def build(cls, scans, binds=True, **kwargs):
        scans = tuple(scans)
        try:
            scan = scans[0]
            if isinstance(scan, ScanBunch):
                def generate():
                    for bunch in scans:
                        yield bunch.precursor
                        for product in bunch.products:
                            yield product
                return cls.build(generate())
            elif isinstance(scan, Scan):
                scans = tuple(s.clone()._load() for s in scans)
            elif isinstance(scan, ProcessedScan):
                scans = tuple(s.clone() for s in scans)
            else:
                raise TypeError("Cannot build an in-memory scan source from {}".format(
                    type(scan)))
        except IndexError:
            raise ValueError("Must pass a non-empty iterable")
        # This alters the index of the scan while it may still be bound
        # to another scan source for attribute lookup. At the time of
        # writing this is needed to properly check bounds on ScanIterator
        # methods, but may lead to problems if the index is needed to resolve
        # other attributes.
        for i, scan in enumerate(scans):
            scan.index = i
        return cls(scans, binds=binds, **kwargs)

    def reset(self):
        self.make_iterator(None)

    def _build_indices(self):
        self._scan_id_map = dict()
        self._scan_index_map = OrderedDict()
        self._index = OrderedDict()

        for scan in self._scans:
            if self._binds:
                scan.bind(self)
            self._scan_id_map[scan.id] = scan
            self._scan_index_map[scan.index] = scan
            self._index[scan.id] = scan.index

    @property
    def index(self):
        return self._index

    def __len__(self):
        return len(self.index)

    def _validate(self, scan):
        return True

    def _make_scan_index_producer(self, start_index=None, start_time=None):
        if start_index is not None:
            return range(start_index, len(self._scans))
        elif start_time is not None:
            start_index = self.get_scan_by_time(start_time).index
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            return range(start_index, len(self._scans))
        else:
            return range(0, len(self._scans))

    def _make_default_iterator(self):
        return self._make_scan_index_producer()

    def next(self):
        return next(self._producer)

    def get_scan_by_id(self, scan_id):
        return self._scan_id_map[scan_id]

    def get_scan_by_time(self, time):
        scan_ids = tuple(self.index)
        lo = 0
        hi = len(scan_ids)
        while hi != lo:
            mid = (hi + lo) // 2
            sid = scan_ids[mid]
            scan = self.get_scan_by_id(sid)
            scan_time = scan.scan_time
            if scan_time == time:
                return scan
            elif (hi - lo) == 1:
                return scan
            elif scan_time > time:
                hi = mid
            else:
                lo = mid

    def get_scan_by_index(self, index):
        return self._scan_index_map[index]

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        if scan_id is not None:
            index = self.get_scan_by_id(scan_id).index
        elif rt is not None:
            index = self.get_scan_by_time(rt).index
        scan = self.get_scan_by_index(index)
        if require_ms1:
            while scan.ms_level > 1:
                index -= 1
                scan = self.get_scan_by_index(index)

        self.make_iterator(self._make_scan_index_producer(index), grouped=grouped)
        return self



import unittest

import numpy as np

from ms_peak_picker import FittedPeak
from brainpy._c.isotopic_distribution import TheoreticalPeak as Peak

from ms_deisotope.scoring import (
        PenalizedMSDeconVFitter, MSDeconVFitter, DotProductFitter, ScaledGTestFitter,
        GTestFitter, LeastSquaresFitter)

experimental = [
    FittedPeak(mz=739.920, intensity=8356.829, signal_to_noise=100.000,
               peak_count=30, index=30588, full_width_at_half_max=0.050, area=425.445),
    FittedPeak(mz=740.255, intensity=8006.456, signal_to_noise=100.000,
               peak_count=31, index=30622, full_width_at_half_max=0.051, area=416.640),
    FittedPeak(mz=740.589, intensity=4970.605, signal_to_noise=100.000,
               peak_count=32, index=30655, full_width_at_half_max=0.050, area=253.103),
    FittedPeak(mz=740.923, intensity=2215.961, signal_to_noise=100.000,
               peak_count=33, index=30688, full_width_at_half_max=0.051, area=118.020)
]

theoretical = [
    Peak(mz=739.920306, intensity=8310.933747, charge=-3),
    Peak(mz=740.254733, intensity=8061.025466, charge=-3),
    Peak(mz=740.588994, intensity=4926.998052, charge=-3),
    Peak(mz=740.923235, intensity=2250.893651, charge=-3)
]


class IsotopicFitScoringTests(unittest.TestCase):

    def test_penalized_msdeconv(self):
        scorer = PenalizedMSDeconVFitter(20, 2.0)
        # score = scorer.evaluate(None, experimental, theoretical)
        scores = [scorer.evaluate(None, experimental, theoretical) for i in range(10)]
        score = scores[0]
        print(scores)
        assert all([np.isclose(s, score) for s in scores[1:]]), scores
        self.assertAlmostEqual(score, 293.47483621051316, 3)
        score = scorer(None, experimental, theoretical)
        self.assertAlmostEqual(score, 293.47483621051316, 3)

    def test_msdeconv(self):
        scorer = MSDeconVFitter()
        score = scorer.evaluate(None, experimental, theoretical)
        self.assertAlmostEqual(score, 293.5135479544602, 3)
        score = scorer(None, experimental, theoretical)
        self.assertAlmostEqual(score, 293.5135479544602, 3)

    def test_dotproduct(self):
        scorer = DotProductFitter()
        score = scorer.evaluate(None, experimental, theoretical)
        self.assertAlmostEqual(score, 163471351.56044182, 3)
        score = scorer(None, experimental, theoretical)
        self.assertAlmostEqual(score, 163471351.56044182, 3)

    def test_scaled_g_test(self):
        scorer = ScaledGTestFitter()
        score = scorer.evaluate(None, experimental, theoretical)
        self.assertAlmostEqual(score, 6.59454124295588e-05, 3)
        score = scorer(None, experimental, theoretical)
        self.assertAlmostEqual(score, 6.59454124295588e-05, 3)

    def test_g_test(self):
        scorer = GTestFitter()
        score = scorer.evaluate(None, experimental, theoretical)
        self.assertAlmostEqual(score, 1.5531726368555354, 3)
        score = scorer(None, experimental, theoretical)
        self.assertAlmostEqual(score, 1.5531726368555354, 3)

    def test_least_squares(self):
        scorer = LeastSquaresFitter()
        score = scorer.evaluate(None, experimental, theoretical)
        self.assertAlmostEqual(score, 7.463484119741042e-05, 3)
        score = scorer(None, experimental, theoretical)
        self.assertAlmostEqual(score, 7.463484119741042e-05, 3)


if __name__ == '__main__':
    unittest.main()



import unittest

from ms_deisotope.data_source.metadata import activation, cv

class TestTerm(unittest.TestCase):
    def test_cv_eq(self):
        assert activation.CID == activation.CID
        assert activation.CID == activation.ActivationInformation("CAD", 55.5).method
        assert activation.CID != activation.HCD



import unittest

import numpy as np

from ms_deisotope.data_source import common, mzml
from ms_deisotope.averagine import peptide

from ms_peak_picker import FittedPeak
from ms_peak_picker.peak_statistics import gaussian_shape


points = [(576.5, 3, 8e4), (862.1, 2, 15e4)]
fwhm = 0.05


def make_profile(points, fwhm):
    peaks = []
    i = 0
    for point in points:
        tid = peptide.isotopic_cluster(point[0], point[1], truncate_after=0.99)
        for tp in tid:
            fp = FittedPeak(tp.mz, tp.intensity * point[2], 0, i, i,
                            fwhm, tp.intensity * point[2])
            peaks.append(fp)
    mz = np.array([0])
    intensity = np.array([0])

    for p in peaks:
        x, y = gaussian_shape(p)
        mz = np.concatenate([mz, [x[0] - 0.0001], x, [x[-1] + 0.0001]])
        intensity = np.concatenate([intensity, [0], y, [0]])
    return mz, intensity


class TestScanMachinery(unittest.TestCase):
    def make_scan(self):
        mz, intensity = make_profile(points, fwhm)
        scan = common.Scan(
            {
                "id": "spam",
                "index": 0,
                "ms level": 1,
                "m/z array": mz,
                "intensity array": intensity,
                "profile spectrum": "",
                "positive scan": "",
                "scanList": {
                    "scan": [
                        {"scan start time": 0}
                    ]
                }
            },
            mzml.MzMLDataInterface())
        return scan

    def make_broken_vendor_scan(self):
        base = self.make_scan()

        class BrokenVendorPeakPickingMzMLDataInterface(mzml.MzMLDataInterface):
            def _pick_peaks_vendor(self, scan, *args, **kwargs):
                raise ValueError("This was a horrible idea.")

        scan = common.Scan(base._data, BrokenVendorPeakPickingMzMLDataInterface())
        return scan

    def test_pick_peaks(self):
        scan = self.make_scan()
        self.assertIsNone(scan.peak_set)
        scan.pick_peaks()
        self.assertIsNotNone(scan.peak_set)
        self.assertEqual(len(scan.peak_set), 10)
        self.assertIsNotNone(scan.has_peak(576.5))

        broken_scan = self.make_broken_vendor_scan()
        assert broken_scan.pick_peaks().peak_set == scan.peak_set
        with self.assertRaises(ValueError):
            broken_scan.pick_peaks('vendor')

    def test_equality(self):
        scan = self.make_scan()
        scan2 = self.make_scan()
        self.assertEqual(scan, scan2)

    def test_raw_arrays(self):
        scan = self.make_scan()
        part = scan.arrays.between_mz(575., 577.)
        assert part.intensity.sum() > 0
        assert (scan.arrays * 2).between_mz(575., 577.).intensity.sum() > part.intensity.sum()


if __name__ == '__main__':
    unittest.main()



import unittest

from ms_deisotope import processor
from ms_deisotope.averagine import glycopeptide, peptide
from ms_deisotope.scoring import PenalizedMSDeconVFitter, MSDeconVFitter

from ms_deisotope.test.common import datafile


class TestScanProcessor(unittest.TestCase):
    mzml_path = datafile("three_test_scans.mzML")
    missing_charge_mzml = datafile("has_missing_charge_state_info.mzML")
    complex_compressed_mzml = datafile("20150710_3um_AGP_001_29_30.mzML.gz")

    def test_processor(self):
        proc = processor.ScanProcessor(self.mzml_path, ms1_deconvolution_args={
            "averagine": glycopeptide,
            "scorer": PenalizedMSDeconVFitter(5., 2.)
        })
        for scan_bunch in iter(proc):
            self.assertIsNotNone(scan_bunch)
            self.assertIsNotNone(scan_bunch.precursor)
            self.assertIsNotNone(scan_bunch.products)

    def test_averaging_processor(self):
        proc = processor.ScanProcessor(self.mzml_path, ms1_deconvolution_args={
            "averagine": glycopeptide,
            "scorer": PenalizedMSDeconVFitter(5., 2.)
        }, ms1_averaging=1)
        for scan_bunch in iter(proc):
            self.assertIsNotNone(scan_bunch)
            self.assertIsNotNone(scan_bunch.precursor)
            self.assertIsNotNone(scan_bunch.products)

    def test_missing_charge_processing(self):
        proc = processor.ScanProcessor(self.missing_charge_mzml, ms1_deconvolution_args={
            "averagine": glycopeptide,
            "scorer": PenalizedMSDeconVFitter(5., 2.)
        })
        for scan_bunch in iter(proc):
            self.assertIsNotNone(scan_bunch)
            self.assertIsNotNone(scan_bunch.precursor)
            self.assertIsNotNone(scan_bunch.products)
            for product in scan_bunch.products:
                if product.precursor_information.defaulted:
                    candidates = scan_bunch.precursor.peak_set.between(
                        product.precursor_information.mz - 1, product.precursor_information.mz + 1)
                    assert len(candidates) == 0

    def test_complex_processor(self):
        proc = processor.ScanProcessor(self.complex_compressed_mzml, ms1_deconvolution_args={
            "averagine": glycopeptide,
            "scorer": PenalizedMSDeconVFitter(20., 2.),
            "truncate_after": 0.95
        }, msn_deconvolution_args={
            "averagine": peptide,
            "scorer": MSDeconVFitter(10.),
            "truncate_after": 0.8
        })
        bunch = next(proc)
        assert len(bunch.products) == 5
        for product in bunch.products:
            assert not product.precursor_information.defaulted
        recalculated_precursors = {
            'scanId=1740086': 4640.00074242012,
            'scanId=1740149': 4786.05878475792,
            'scanId=1740226': 4640.007868154431,
            'scanId=1740344': 4348.90894554512,
            'scanId=1740492': 5005.1329902247435
        }
        for product in bunch.products:
            mass = product.precursor_information.extracted_neutral_mass
            self.assertAlmostEqual(mass, recalculated_precursors[product.id], 2)

        proc.start_from_scan("scanId=1760847")
        bunch = next(proc)
        recalculated_precursors = {
            'scanId=1761168': 4640.01972225792,
            'scanId=1761235': 4640.019285920238,
            'scanId=1761325': 4786.07251976387,
            'scanId=1761523': 4696.016295197582,
            'scanId=1761804': 986.58798612896
        }
        for product in bunch.products:
            mass = product.precursor_information.extracted_neutral_mass
            self.assertAlmostEqual(mass, recalculated_precursors[product.id], 2)


if __name__ == '__main__':
    unittest.main()



import unittest
import os

from ms_deisotope.data_source import MzMLLoader
from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type

scan_ids = [
    "controllerType=0 controllerNumber=1 scan=10014",
    "controllerType=0 controllerNumber=1 scan=10015",
    "controllerType=0 controllerNumber=1 scan=10016"
]


class TestMzMLLoaderScanBehavior(unittest.TestCase):
    path = datafile("three_test_scans.mzML")
    only_ms2_path = datafile("only_ms2_mzml.mzML")

    @property
    def reader(self):
        reader = infer_type.MSFileLoader(self.path)
        assert len(reader.index) == 3
        assert reader.index.from_index(0) == scan_ids[0]
        assert list(reader.index.index_sequence) == sorted(
            reader.index.index_sequence, key=lambda x: x[1])
        return reader

    def test_index_building(self):
        MzMLLoader.prebuild_byte_offset_file(self.path)
        parser = MzMLLoader._parser_cls(self.path)
        assert parser._check_has_byte_offset_file()
        index = parser.index
        offsets = index['spectrum']
        key_list = list(offsets.keys())
        assert key_list == scan_ids
        offset_file_name = parser._byte_offset_filename
        try:
            os.remove(offset_file_name)
        except OSError:
            pass

    def test_index_integrity(self):
        reader = self.reader
        reader.make_iterator(grouped=False)
        for i, scan in enumerate(reader):
            assert i == scan.index

    def test_iteration(self):
        reader = self.reader
        i = 0
        bunch = next(reader)
        if bunch.precursor:
            i += 1
        i += len(bunch.products)
        self.assertEqual(i, 3)

        reader.reset()
        reader.make_iterator(grouped=False)

        scan = next(reader)
        scan._load()
        self.assertEqual(scan.index, 0)

        reader.close()

    def test_index(self):
        reader = self.reader
        bunch = next(reader)
        self.assertEqual(bunch.precursor.index, 0)
        for i, scan in enumerate(bunch.products, 1):
            self.assertEqual(scan.index, i)
        reader.close()

    def test_ms_level(self):
        reader = self.reader
        bunch = next(reader)
        self.assertEqual(bunch.precursor.ms_level, 1)
        for i, scan in enumerate(bunch.products, 1):
            self.assertEqual(scan.ms_level, 2)
        reader.close()

    def test_polarity(self):
        reader = self.reader
        bunch = next(reader)
        self.assertEqual(bunch.precursor.polarity, 1)
        reader.close()

    def test_activation(self):
        reader = self.reader
        bunch = next(reader)
        self.assertEqual(bunch.precursor.activation, None)
        for product in bunch.products:
            self.assertNotEqual(product.activation, None)
            self.assertEqual(product.activation.method, "beam-type collision-induced dissociation")
        reader.close()

    def test_precursor(self):
        reader = self.reader
        bunch = next(reader)
        self.assertEqual(bunch.precursor.precursor_information, None)
        for product in bunch.products:
            self.assertNotEqual(product.precursor_information, None)
            self.assertEqual(product.precursor_information.precursor, bunch.precursor)
        reader.close()

    def test_pick_peaks(self):
        reader = self.reader
        bunch = next(reader)
        scan = bunch.precursor.pick_peaks()
        self.assertEqual(len(scan.peak_set), 2108)
        reader.close()

    def test_pack(self):
        reader = self.reader
        bunch = next(reader)
        bunch.precursor.pick_peaks()
        self.assertEqual(bunch.precursor.pack().title, bunch.precursor.title)
        reader.close()

    def test_get_scan_by_id(self):
        reader = self.reader
        precursor = reader.get_scan_by_id(scan_ids[0])
        self.assertEqual(precursor.id, scan_ids[0])
        self.assertEqual(precursor.index, 0)

        product = reader.get_scan_by_id(scan_ids[2])
        self.assertEqual(product.id, scan_ids[2])
        self.assertEqual(product.index, 2)

        self.assertEqual(product.precursor_information.precursor_scan_id, scan_ids[0])
        self.assertIs(precursor, reader.get_scan_by_id(scan_ids[0]))
        reader.close()

    def test_get_scan_by_index(self):
        reader = self.reader

        precursor = reader.get_scan_by_index(0)
        self.assertEqual(precursor.index, 0)
        self.assertEqual(precursor.id, scan_ids[0])
        reader.close()

    def test_get_scan_by_time(self):
        reader = self.reader
        precursor = reader.get_scan_by_time(22.12829)
        self.assertEqual(precursor.id, scan_ids[0])

        product = reader.get_scan_by_time(22.132753)
        self.assertEqual(product.index, 1)
        scan = reader.get_scan_by_time(float('inf'))
        assert scan == reader[-1]
        reader.close()

    def test_instrument_configuration(self):
        reader = self.reader
        bunch = next(reader)
        precursor = bunch.precursor
        config = precursor.instrument_configuration
        self.assertEqual(config.id, "IC1")
        assert "orbitrap" in config.analyzers

    def test_file_description(self):
        file_info = self.reader.file_description()
        assert "MS1 spectrum" in file_info.contents
        assert "MSn spectrum" in file_info.contents
        source_file = file_info.source_files[0]
        assert source_file.name == "three_test_scans.mzML"
        assert "location" not in source_file.parameters

    def test_sample_list(self):
        ms2_reader = MzMLLoader(self.only_ms2_path)
        samples = ms2_reader.samples()
        sample = samples[0]
        assert sample.name == ""
        assert sample['SampleDescription'] == "REACTION VOLT_-0.2 _Mirror RF_125V Reagent accu_150 ms accu time_100 ms"

    def test_acquisition_information(self):
        reader = self.reader
        bunch = next(reader)
        precursor = bunch.precursor
        acquisition = precursor.acquisition_information
        self.assertTrue(
            abs(acquisition[0].start_time - precursor.scan_time) < 1e-3)
        self.assertEqual(len(acquisition), 1)
        window = acquisition[0].total_scan_window()
        if window:
            self.assertIn(precursor.arrays.mz[len(precursor.arrays.mz) // 2], window)

    def test_annotations(self):
        reader = self.reader
        bunch = next(reader)
        precursor = bunch.precursor
        assert len(precursor.annotations) > 0

    def test_iteration_mode_detection(self):
        reader = infer_type.MSFileLoader(self.only_ms2_path)
        assert reader.iteration_mode == 'single'

    def test_source_file_parsing(self):
        reader = self.reader
        finfo = reader.file_description()
        sf = finfo.source_files[0]
        assert sf.name == 'three_test_scans.mzML'
        assert isinstance(sf.path, str)

        reader = infer_type.MSFileLoader(self.only_ms2_path)
        finfo = reader.file_description()
        sf = finfo.source_files[0]
        assert sf.name == 'analysis.baf'
        assert isinstance(sf.path, str)

    def test_data_processing_parsing(self):
        reader = infer_type.MSFileLoader(self.only_ms2_path)
        assert len(reader.data_processing()[0]) == 3

    def test_software_list(self):
        reader = infer_type.MSFileLoader(self.path)
        assert len(reader.software_list()) == 2


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .lcms_feature import LCMSFeature


def peak_indices(x, min_height=0):
    """Find the index of local maxima.

    Parameters
    ----------
    x : np.ndarray
        Data to find local maxima in
    min_height : float, optional
        Minimum peak height

    Returns
    -------
    np.ndarray[int]
        Indices of maxima in x

    References
    ----------
    https://github.com/demotu/BMC/blob/master/functions/detect_peaks.py
    """
    if x.size < 3:
        return np.array([], dtype=int)
    # find indices of all peaks
    dx = x[1:] - x[:-1]
    rising_edges = np.where((np.hstack((dx, 0)) <= 0) &
                            (np.hstack((0, dx)) > 0))[0]
    falling_edges = np.where((np.hstack((dx, 0)) < 0) &
                             (np.hstack((0, dx)) >= 0))[0]
    indices = np.unique(np.hstack((rising_edges, falling_edges)))
    if indices.size and min_height > 0:
        indices = indices[x[indices] >= min_height]
    return indices


class ValleyPoint(object):
    __slots__ = ["first_maximum", "minimum", "second_maximum", "first_maximum_index",
                 "minimum_index", "second_maximum_index", "total_distance"]

    def __init__(self, first_maximum, minimum, second_maximum, first_maximum_index,
                 minimum_index, second_maximum_index):
        self.first_maximum = first_maximum
        self.minimum = minimum
        self.second_maximum = second_maximum
        self.first_maximum_index = first_maximum_index
        self.minimum_index = minimum_index
        self.second_maximum_index = second_maximum_index
        self.total_distance = self.compute_distance()

    def compute_distance(self):
        return (self.first_maximum - self.minimum) + (self.second_maximum - self.minimum)

    def __repr__(self):
        return "ValleyPoint(%0.2f, %0.2f, %0.2f, %0.3e)" % (
            self.first_maximum_index, self.minimum_index, self.second_maximum_index, self.total_distance)

    def split(self, profile):
        index = profile.nodes.find_time(self.minimum_index)[1]
        before = LCMSFeature(profile[:index])
        after = LCMSFeature(profile[index:])
        return before, after


class PeakBoundary(object):
    def __init__(self, first_minimum, maximum, second_minimum,
                 first_minimum_index, maximum_index, second_minimum_index):
        self.first_minimum = first_minimum
        self.maximum = maximum
        self.second_minimum = second_minimum
        self.first_minimum_index = first_minimum_index
        self.maximum_index = maximum_index
        self.second_minimum_index = second_minimum_index
        self.total_distance = self.compute_distance()

    def compute_distance(self):
        return (self.maximum - self.first_minimum) + (self.maximum - self.second_minimum)

    def __repr__(self):
        return "PeakBoundary(%0.2f, %0.2f, %0.2f, %0.3e)" % (
            self.first_minimum_index, self.maximum_index, self.second_minimum_index,
            self.total_distance)

    def split(self, profile):
        first_minimum = profile.nodes.find_time(self.first_minimum_index)[1]
        second_minimum = profile.nodes.find_time(self.second_minimum_index)[1]
        if first_minimum == 0:
            first_minimum += 1
        prefix = profile[:first_minimum]
        peak = profile[(first_minimum - 1):second_minimum + 1]
        suffix = profile[second_minimum + 1:]
        cases = LCMSFeature(prefix), LCMSFeature(peak), LCMSFeature(suffix)
        return cases


def interpolate(xs, ys, n=200):
    new_xs = np.linspace(xs.min(), xs.max(), n)
    new_ys = np.interp(new_xs, xs, ys)
    return new_xs, new_ys


def sliding_mean(ys):
    return (np.concatenate((ys[1:], [0])) + ys + np.concatenate(([0], ys[:-1]))) / 3


def sliding_median(ys):
    arr = np.zeros_like(ys)
    n = len(ys)
    for i in range(n):
        if i == 0:
            value = np.median(np.concatenate(([0], ys[:2]),))
        elif i == n - 1:
            value = np.median(np.concatenate(([0], ys[n - 2:]),))
        else:
            value = np.median(ys[i - 1:i + 2])
        arr[i] = value
    return arr


def gauss(x, mean, variance, scale=2):
    return 1. / np.sqrt(2 * np.pi * variance) * np.exp(-(x - mean) ** 2 / (2 * variance))


def binsearch(array, value):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        point = array[mid]
        if value == point:
            return mid
        elif hi - lo == 1:
            return mid
        elif point > value:
            hi = mid
        else:
            lo = mid


def gaussian_smooth(x, y, width=0.05):
    smoothed = np.zeros_like(y)

    n = x.shape[0]
    for i in range(n):
        low_edge = binsearch(x, x[i] - width)
        high_edge = binsearch(x, x[i] + width)
        if high_edge - 1 == low_edge or high_edge == low_edge:
            smoothed[i] = y[i]
            continue
        x_slice = x[low_edge:high_edge]
        y_slice = y[low_edge:high_edge]
        center = np.mean(x_slice)
        spread = np.var(x_slice)
        weights = gauss(x_slice, center, spread)
        val = ((y_slice * weights).sum() / weights.sum())
        if np.isnan(val):
            raise ValueError("NaN")
        smoothed[i] = val
    return smoothed


def smooth_leveled(xs, ys, level=0):
    if level == 0:
        return ys
    elif level == 1:
        return sliding_mean(ys)
    elif level == 2:
        return sliding_mean(sliding_median(ys))
    elif level == 3:
        return gaussian_smooth(xs, ys, 0.05)
    else:
        return gaussian_smooth(xs, ys, level)


class ProfileSplitter(object):
    def __init__(self, profile):
        self.profile = profile
        self.xs, self.ys = profile.as_arrays()
        self.partition_sites = []

    def _extreme_indices(self, ys):
        maxima_indices = peak_indices(ys)
        minima_indices = peak_indices(-ys)
        return maxima_indices, minima_indices

    def interpolate(self, xs, ys, n=200):
        new_xs = np.linspace(xs.min(), xs.max(), n)
        new_ys = np.interp(new_xs, xs, ys)
        return new_xs, new_ys

    def locate_peak_boundaries(self, smooth=1, interpolate_past=200):
        xs = self.xs
        ys = self.ys
        if len(xs) > interpolate_past:
            xs, ys = self.interpolate(xs, ys, interpolate_past)
        if smooth:
            ys = smooth_leveled(xs, ys, smooth)

        maxima_indices, minima_indices = self._extreme_indices(ys)
        candidates = []
        for i in range(len(minima_indices)):
            min_i = minima_indices[i]
            for j in range(i + 1, len(minima_indices)):
                min_j = minima_indices[j]
                for k in range(len(maxima_indices)):
                    max_k = maxima_indices[k]
                    y_i = ys[min_i]
                    y_j = ys[min_j]
                    y_k = ys[max_k]
                    if min_i < max_k < min_j and (y_k - y_i) > (y_k * 0.01) and (
                            y_k - y_j) > (y_k * 0.01):
                        point = PeakBoundary(y_i, y_k, y_j, xs[min_i], xs[max_k], xs[min_j])
                        candidates.append(point)
        if candidates:
            candidates = sorted(candidates, key=lambda x: x.total_distance, reverse=True)
            best_point = candidates[0]
            self.partition_sites.append(best_point)

        return candidates

    def locate_valleys(self, scale=0.3, smooth=1, interpolate_past=200):
        xs = self.xs
        ys = self.ys

        if len(xs) > interpolate_past:
            xs, ys = self.interpolate(xs, ys, interpolate_past)

        if smooth:
            ys = smooth_leveled(xs, ys, smooth)

        maxima_indices, minima_indices = self._extreme_indices(ys)
        candidates = []

        for i in range(len(maxima_indices)):
            max_i = maxima_indices[i]
            for j in range(i + 1, len(maxima_indices)):
                max_j = maxima_indices[j]
                for k in range(len(minima_indices)):
                    min_k = minima_indices[k]
                    y_i = ys[max_i]
                    y_j = ys[max_j]
                    y_k = ys[min_k]
                    if max_i < min_k < max_j and (y_i - y_k) > (y_i * scale) and (
                            y_j - y_k) > (y_j * scale):
                        point = ValleyPoint(y_i, y_k, y_j, xs[max_i], xs[min_k], xs[max_j])
                        candidates.append(point)
        if candidates:
            candidates = sorted(candidates, key=lambda x: x.total_distance, reverse=True)
            best_point = candidates[0]
            self.partition_sites.append(best_point)

        return candidates

    def extract_peak(self, peak_boundary):
        cases = peak_boundary.split(self.profile)
        return cases

    def split_valley(self, valley):
        return valley.split(self.profile)


def split_valleys(profiles, scale=0.3, n_levels=2):
    for i in range(n_levels):
        out = []
        for case in profiles:
            ps = ProfileSplitter(case)
            vallies = ps.locate_valleys(scale)
            if vallies:
                splitter = vallies[0]
                out.extend(p for p in ps.split_valley(splitter) if len(p) > 0)
            else:
                out.append(case)
        profiles = out
    return profiles


def exhaustive_split(profiles, scale=0.5, maxiter=100, smooth=.15):
    hold = []
    for i in range(maxiter):
        n = len(profiles) + len(hold)
        out = []
        for case in profiles:
            ps = ProfileSplitter(case)
            vallies = ps.locate_valleys(scale, smooth=smooth)
            if vallies:
                splitter = vallies[0]
                out.extend(p for p in ps.split_valley(splitter) if len(p) > 0)
            else:
                hold.append(case)
        profiles = out
        if n == len(profiles) + len(hold):
            break
    else:
        print("Did not converge in %d iterations (%d != %d)" % (i, n, len(profiles) + len(hold)))
    return profiles + hold


def is_flat(feature, fraction=0.5, smooth=1):
    xs, ys = feature.as_arrays()
    ys = smooth_leveled(xs, ys, smooth)
    apex = ys.max()
    return apex * fraction < ys[0] and apex * fraction < ys[-1]


def clean_profiles(features, scale=0.5, maxiter=100, smooth=0.15):
    features = [c for f in features for c in f.split_sparse(0.25)]
    features = exhaustive_split(features, scale=scale, smooth=smooth)
    features = [fe for fe in features if len(fe) > 2 and not is_flat(
                fe, fraction=0.8, smooth=3)]
    return features


try:
    has_c = True
    _ProfileSplitter = ProfileSplitter
    _ValleyPoint = ValleyPoint
    _split_valleys = split_valleys
    _smooth_leveled = smooth_leveled
    _gaussian_smooth = gaussian_smooth

    from ms_deisotope._c.feature_map.profile_transform import (
        ProfileSplitter, ValleyPoint, split_valleys,
        gaussian_smooth, smooth_leveled)
except ImportError:
    has_c = False



import unittest
import os
import tempfile
import gzip

import ms_deisotope
from ms_deisotope.data_source import MzMLLoader
from ms_deisotope.data_source import _compression
from ms_deisotope.test.common import datafile

from ms_deisotope.output.mzml import MzMLSerializer, ProcessedMzMLDeserializer


class TestMzMLSerializer(unittest.TestCase):
    source_data_path = datafile("three_test_scans.mzML")

    def test_writer(self):
        source_reader = MzMLLoader(self.source_data_path)
        fd, name = tempfile.mkstemp()
        with open(name, 'wb') as fh:
            writer = MzMLSerializer(fh, n_spectra=len(source_reader.index), deconvoluted=True)
            description = source_reader.file_description()
            writer.add_file_information(description)
            writer.add_file_contents("profile spectrum")
            writer.add_file_contents("centroid spectrum")
            writer.remove_file_contents("profile spectrum")

            instrument_configs = source_reader.instrument_configuration()
            for config in instrument_configs:
                writer.add_instrument_configuration(config)

            software_list = source_reader.software_list()
            for software in software_list:
                writer.add_software(software)

            data_processing_list = source_reader.data_processing()
            for dp in data_processing_list:
                writer.add_data_processing(dp)

            processing = writer.build_processing_method()
            writer.add_data_processing(processing)
            bunch = next(source_reader)
            bunch.precursor.pick_peaks()
            bunch.precursor.deconvolute()
            for product in bunch.products:
                product.pick_peaks()
                product.deconvolute()
            writer.save(bunch)
            writer.complete()
            fh.flush()
            writer.format()
        source_reader.reset()
        processed_reader = ProcessedMzMLDeserializer(_compression.get_opener(writer.handle.name))

        for a, b in zip(source_reader.instrument_configuration(), processed_reader.instrument_configuration()):
            assert a.analyzers == b.analyzers
        for a, b in zip(source_reader, processed_reader):
            assert a.precursor.id == b.precursor.id
            assert (a.precursor.acquisition_information == b.precursor.acquisition_information)
            for an, bn in zip(a.products, b.products):
                assert an.id == bn.id
                assert abs(an.precursor_information.neutral_mass - bn.precursor_information.neutral_mass) < 1e-6
        processed_reader.reset()
        description = processed_reader.file_description()
        assert "profile spectrum" not in description.contents
        assert "centroid spectrum" in description.contents
        sf = description.source_files[0]
        assert 'location' not in sf.parameters
        assert sf.parameters['SHA-1'] == 'a2a091b82f27676da87a6c7d17cc90d2d90b8fbf'
        assert processed_reader.sample_run.name == "sample_1"
        index = processed_reader.extended_index
        pinfo = index.find_msms_by_precursor_mass(ms_deisotope.neutral_mass(562.7397, 2))
        assert len(pinfo) > 0

        processed_reader.close()
        try:
            os.remove(name)
            os.remove(processed_reader._index_file_name)
        except OSError:
            pass


if __name__ == '__main__':
    unittest.main()
